#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import importlib
import logging
from operator import itemgetter
from operator import attrgetter
import os
import pkgutil

from libnmstate import validator
from libnmstate.error import NmstateError
from libnmstate.error import NmstateValueError
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .nispor.plugin import NisporPlugin
from .plugin import NmstatePlugin
from .state import merge_dict

_INFO_TYPE_RUNNING = 1
_INFO_TYPE_RUNNING_CONFIG = 2
_INFO_TYPE_SAVED_CONFIG = 3


@contextmanager
def plugin_context():
    plugins = _load_plugins()
    try:
        # Lowest priority plugin should perform actions first.
        plugins.sort(key=attrgetter("priority"))
        yield plugins
    except (Exception, KeyboardInterrupt):
        for plugin in plugins:
            if plugin.checkpoint:
                try:
                    plugin.rollback_checkpoint()
                # Don't complex thing by raise exception when handling another
                # exception, just log the rollback failure.
                except Exception as e:
                    logging.error(f"Rollback failed with error {e}")
        raise
    finally:
        for plugin in plugins:
            plugin.unload()


def show_with_plugins(
    plugins, include_status_data=None, info_type=_INFO_TYPE_RUNNING
):
    for plugin in plugins:
        plugin.refresh_content()
    report = {}
    if include_status_data and info_type == _INFO_TYPE_RUNNING:
        report["capabilities"] = plugins_capabilities(plugins)

    report[Interface.KEY] = _get_interface_info_from_plugins(
        plugins, info_type
    )

    report[Route.KEY] = _get_routes_from_plugins(plugins, info_type)

    report[RouteRule.KEY] = _get_route_rules_from_plugins(plugins, info_type)

    dns_plugin = _find_plugin_for_capability(
        plugins, NmstatePlugin.PLUGIN_CAPABILITY_DNS
    )
    if dns_plugin:
        if info_type == _INFO_TYPE_SAVED_CONFIG:
            report[DNS.KEY] = {
                DNS.CONFIG: dns_plugin.get_saved_dns_client_config()
            }
        else:
            report[DNS.KEY] = dns_plugin.get_dns_client_config()
            if info_type != _INFO_TYPE_RUNNING:
                report[DNS.KEY].pop(DNS.RUNNING, None)

    for plugin in plugins:
        report.update(plugin.get_global_state())

    validator.schema_validate(report)
    return report


def plugins_capabilities(plugins):
    capabilities = set()
    for plugin in plugins:
        capabilities.update(set(plugin.capabilities))
    return list(capabilities)


def _load_plugins():
    plugins = [NisporPlugin()]
    plugins.extend(_load_external_py_plugins())
    plugins.extend(_load_nm_plugin())
    return plugins


def _load_nm_plugin():
    """
    Makin NetworkManager plugin as optional
    """
    try:
        from libnmstate.nm import NetworkManagerPlugin

        return [NetworkManagerPlugin()]
    except Exception as e:
        logging.warning(f"Failed to load NetworkManager plugin: {e}")
        return []


def _load_external_py_plugins():
    """
    Load module from folder defined in system evironment NMSTATE_PLUGIN_DIR,
    if empty, use the 'plugins' folder of current python file.
    """
    plugins = []
    plugin_dir = os.environ.get("NMSTATE_PLUGIN_DIR")
    if not plugin_dir:
        plugin_dir = f"{os.path.dirname(os.path.realpath(__file__))}/plugins"

    for _, name, ispkg in pkgutil.iter_modules([plugin_dir]):
        if name.startswith("nmstate_plugin_"):
            try:
                spec = importlib.util.spec_from_file_location(
                    name, f"{plugin_dir}/{name}.py"
                )
                plugin_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(plugin_module)
                plugin = plugin_module.NMSTATE_PLUGIN()
                plugins.append(plugin)
            except Exception as error:
                logging.warning(f"Failed to load plugin {name}: {error}")

    return plugins


def _find_plugin_for_capability(plugins, capability):
    """
    Return the plugin with specified capability and highest priority.
    """
    chose_plugin = None
    for plugin in plugins:
        if capability in plugin.plugin_capabilities:
            if (
                chose_plugin and plugin.priority > chose_plugin.priority
            ) or not chose_plugin:
                chose_plugin = plugin
    return chose_plugin


def _get_interface_info_from_plugins(plugins, info_type):
    all_ifaces = {}
    IFACE_PRIORITY_METADATA = "_plugin_priority"
    IFACE_PLUGIN_SRC_METADATA = "_plugin_source"
    for plugin in plugins:
        if (
            NmstatePlugin.PLUGIN_CAPABILITY_IFACE
            not in plugin.plugin_capabilities
        ):
            continue
        if info_type == _INFO_TYPE_RUNNING_CONFIG:
            ifaces = plugin.get_running_config_interfaces()
        elif info_type == _INFO_TYPE_SAVED_CONFIG:
            ifaces = plugin.get_saved_config_interfaces()
        else:
            ifaces = plugin.get_interfaces()
        for iface in ifaces:
            iface[IFACE_PRIORITY_METADATA] = plugin.priority
            iface[IFACE_PLUGIN_SRC_METADATA] = [plugin.name]
            iface_name = iface[Interface.NAME]
            iface_type = iface.get(Interface.TYPE, InterfaceType.UNKNOWN)
            iface_index = f"{iface_type}.{iface_name}"
            if iface_index in all_ifaces:
                existing_iface = all_ifaces[iface_index]
                existing_priority = existing_iface[IFACE_PRIORITY_METADATA]
                current_priority = plugin.priority
                if current_priority > existing_priority:
                    merge_dict(iface, existing_iface)
                    iface[IFACE_PLUGIN_SRC_METADATA].extend(
                        existing_iface[IFACE_PLUGIN_SRC_METADATA]
                    )
                    all_ifaces[iface_index] = iface
                else:
                    merge_dict(existing_iface, iface)
                    existing_iface[IFACE_PLUGIN_SRC_METADATA].extend(
                        iface[IFACE_PLUGIN_SRC_METADATA]
                    )
            else:
                all_ifaces[iface_index] = iface

    # For any unknown interface, we merge it when there is only one interface
    # with the same name and valid type.
    # This save plugin(e.g. ovsdb) from detecting interface type but still want
    # data been merged.
    to_be_removed_index = set()
    for iface in all_ifaces.values():
        iface_type = iface.get(Interface.TYPE, InterfaceType.UNKNOWN)
        iface_name = iface[Interface.NAME]
        if iface_type == InterfaceType.UNKNOWN:
            iface_types = _get_iface_types_by_name(
                all_ifaces.values(), iface_name
            )
            if len(iface_types) == 1:
                to_be_removed_index.add(
                    f"{InterfaceType.UNKNOWN}.{iface_name}"
                )
                iface_index = f"{iface_types[0]}.{iface_name}"
                existing_iface = all_ifaces[iface_index]
                iface[Interface.TYPE] = existing_iface[Interface.TYPE]
                current_priority = iface[IFACE_PRIORITY_METADATA]
                existing_priority = existing_iface[IFACE_PRIORITY_METADATA]
                if current_priority > existing_priority:
                    merge_dict(iface, existing_iface)
                    iface[IFACE_PLUGIN_SRC_METADATA].extend(
                        existing_iface[IFACE_PLUGIN_SRC_METADATA]
                    )
                    all_ifaces[iface_index] = iface
                else:
                    merge_dict(existing_iface, iface)
                    existing_iface[IFACE_PLUGIN_SRC_METADATA].extend(
                        iface[IFACE_PLUGIN_SRC_METADATA]
                    )

    supplemental_plugin_names = set(
        [plugin.name for plugin in plugins if plugin.is_supplemental_only]
    )

    for iface_index, iface in all_ifaces.items():
        cur_plugin_names = set(iface[IFACE_PLUGIN_SRC_METADATA])
        if cur_plugin_names.issubset(supplemental_plugin_names):
            to_be_removed_index.add(iface_index)

    for iface_index in to_be_removed_index:
        del all_ifaces[iface_index]

    # Remove metadata
    for iface in all_ifaces.values():
        iface.pop(IFACE_PRIORITY_METADATA)
        iface.pop(IFACE_PLUGIN_SRC_METADATA)

    return sorted(all_ifaces.values(), key=itemgetter(Interface.NAME))


def create_checkpoints(plugins, timeout):
    """
    Return a string containing all the check point created by each plugin in
    the format:
        plugin.name|<checkpoing_path>|plugin.name|<checkpoing_path|...

    """
    checkpoints = []
    for plugin in plugins:
        checkpoint = plugin.create_checkpoint(timeout)
        if checkpoint:
            checkpoints.append(f"{plugin.name}|{checkpoint}")
    return "|".join(checkpoints)


def destroy_checkpoints(plugins, checkpoints):
    _checkpoint_action(plugins, _parse_checkpoints(checkpoints), "destroy")


def rollback_checkpoints(plugins, checkpoints):
    _checkpoint_action(plugins, _parse_checkpoints(checkpoints), "rollback")


def _checkpoint_action(plugins, checkpoint_index, action):
    errors = []
    for plugin in plugins:
        if checkpoint_index and plugin.name not in checkpoint_index:
            continue
        checkpoint = (
            checkpoint_index[plugin.name] if checkpoint_index else None
        )
        try:
            if action == "destroy":
                plugin.destroy_checkpoint(checkpoint)
            else:
                plugin.rollback_checkpoint(checkpoint)
        except (Exception, KeyboardInterrupt) as error:
            errors.append(error)

    if errors:
        if len(errors) == 1:
            raise errors[0]
        else:
            raise NmstateError(
                "Got multiple exception during checkpoint "
                f"{action}: {errors}"
            )


def _parse_checkpoints(checkpoints):
    """
    Return a dict mapping plugin name to checkpoint
    """
    if not checkpoints:
        return None
    parsed = checkpoints.split("|")
    if len(parsed) % 2:
        raise NmstateValueError("Invalid format of checkpoint")
    checkpoint_index = {}
    for plugin_name, checkpoint in zip(parsed[0::2], parsed[1::2]):
        checkpoint_index[plugin_name] = checkpoint


def _get_routes_from_plugins(plugins, info_type):
    ret = {Route.RUNNING: [], Route.CONFIG: []}
    for plugin in plugins:
        if NmstatePlugin.PLUGIN_CAPABILITY_ROUTE in plugin.plugin_capabilities:
            if info_type == _INFO_TYPE_SAVED_CONFIG:
                plugin_routes = plugin.get_saved_routes()
                ret[Route.CONFIG].extend(plugin_routes)
            else:
                plugin_routes = plugin.get_routes()
                if info_type == _INFO_TYPE_RUNNING:
                    ret[Route.RUNNING].extend(
                        plugin_routes.get(Route.RUNNING, [])
                    )
                ret[Route.CONFIG].extend(plugin_routes.get(Route.CONFIG, []))
    if info_type != _INFO_TYPE_RUNNING:
        ret.pop(Route.RUNNING)
    return ret


def _get_route_rules_from_plugins(plugins, info_type):
    ret = {RouteRule.CONFIG: []}
    for plugin in plugins:
        if (
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE_RULE
            in plugin.plugin_capabilities
        ):
            if info_type == _INFO_TYPE_SAVED_CONFIG:
                ret[RouteRule.CONFIG].extend(plugin.get_saved_route_rules())
            else:
                plugin_route_rules = plugin.get_route_rules()
                ret[RouteRule.CONFIG].extend(
                    plugin_route_rules.get(RouteRule.CONFIG, [])
                )
    return ret


def _get_iface_types_by_name(iface_infos, name):
    """
    Return the type of all ifaces with valid type and specified name
    """
    iface_types = []
    for iface_info in iface_infos:
        iface_type = iface_info.get(Interface.TYPE, InterfaceType.UNKNOWN)
        if (
            iface_type != InterfaceType.UNKNOWN
            and iface_info[Interface.NAME] == name
        ):
            iface_types.append(iface_type)

    return iface_types


def show_running_config_with_plugins(plugins):
    return show_with_plugins(plugins, info_type=_INFO_TYPE_RUNNING_CONFIG)


def show_saved_config_with_plugins(plugins):
    return show_with_plugins(plugins, info_type=_INFO_TYPE_SAVED_CONFIG)



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time

import pytest

from libnmstate.nm.checkpoint import CheckPoint
from libnmstate.nm.checkpoint import get_checkpoints
from libnmstate.error import NmstateConflictError
from libnmstate.error import NmstateValueError


def test_creating_one_checkpoint(nm_context):
    """ I can create a checkpoint """
    checkpoint = CheckPoint.create(nm_context)
    assert checkpoint is not None
    checkpoint.destroy()


def test_creating_two_checkpoints(nm_context):
    """ I cannot create a checkpoint when a checkpoint already exists. """
    checkpoint = CheckPoint.create(nm_context)
    with pytest.raises(NmstateConflictError):
        CheckPoint.create(nm_context)
    assert checkpoint is not None
    checkpoint.destroy()


def test_checkpoint_timeout(nm_context):
    """ I can create a checkpoint that is removed after one second. """
    checkpoint_a = CheckPoint.create(nm_context, timeout=1)
    time.sleep(1)
    checkpoint_b = CheckPoint.create(nm_context)

    assert checkpoint_b is not None
    assert checkpoint_a is not None
    checkpoint_b.destroy()


def test_getting_a_checkpoint(nm_context):
    """ I can get a list of all checkpoints. """

    checkpoints = get_checkpoints(nm_context.client)

    assert len(checkpoints) == 0

    checkpoint = CheckPoint.create(nm_context)

    nm_context.refresh_content()
    checkpoints = get_checkpoints(nm_context.client)

    assert len(checkpoints) == 1
    assert checkpoints[0] == str(checkpoint)
    checkpoint.destroy()


def test_creating_a_checkpoint_from_dbuspath(nm_context):
    initial_checkpoint = CheckPoint.create(nm_context)
    new_checkpoint = CheckPoint(nm_context, dbuspath=str(initial_checkpoint))
    new_checkpoint.destroy()
    assert not get_checkpoints(nm_context.client)


def test_repeat_create_destroy_checkpoint(nm_context):
    for _ in range(0, 1000):
        checkpoint = CheckPoint.create(nm_context)
        checkpoint.destroy()


def test_plugin_load_specific_checkpoint_not_exist(nm_plugin):
    plugin = nm_plugin
    plugin.create_checkpoint()

    with pytest.raises(NmstateValueError):
        plugin.rollback_checkpoint("checkpoint_not_exist")


def test_plugin_load_default_checkpoint(nm_plugin):
    plugin = nm_plugin
    plugin.create_checkpoint()
    plugin.rollback_checkpoint()

    checkpoints = get_checkpoints(nm_plugin.context.client)

    assert plugin.checkpoint is None
    assert len(checkpoints) == 0


def test_plugin_load_default_checkpoint_when_none_avaiable(nm_plugin):
    plugin = nm_plugin
    with pytest.raises(NmstateValueError):
        plugin.rollback_checkpoint()


def test_plugin_create_and_destroy_checkpoint(nm_plugin):
    plugin = nm_plugin
    plugin.create_checkpoint()
    plugin.destroy_checkpoint()

    checkpoints = get_checkpoints(nm_plugin.context.client)
    assert len(checkpoints) == 0


def test_plugin_create_and_rollback_checkpoint(nm_plugin):
    plugin = nm_plugin
    plugin.create_checkpoint()
    plugin.rollback_checkpoint()

    checkpoints = get_checkpoints(nm_plugin.context.client)
    assert len(checkpoints) == 0


def test_plugin_create_checkpoint_and_timeout(nm_plugin):
    plugin = nm_plugin
    plugin.create_checkpoint(timeout=1)
    time.sleep(2)

    nm_plugin.refresh_content()
    checkpoints = get_checkpoints(nm_plugin.context.client)
    assert len(checkpoints) == 0



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from unittest import mock

from libnmstate import nm


@pytest.fixture
def NM_mock():
    with mock.patch.object(nm.connection, "NM") as m:
        yield m


@pytest.fixture()
def context_mock():
    yield mock.MagicMock()


def test_create_setting(NM_mock):
    con_setting = nm.connection._ConnectionSetting()
    con_setting.create("con-name", "iface-name", "iface-type")

    assert con_setting.setting.props.id == "con-name"
    assert con_setting.setting.props.interface_name == "iface-name"
    assert con_setting.setting.props.uuid
    assert con_setting.setting.props.type == "iface-type"
    assert con_setting.setting.props.autoconnect is True
    assert con_setting.setting.props.autoconnect_slaves == (
        NM_mock.SettingConnectionAutoconnectSlaves.YES
    )


def test_set_controller_setting():
    con_setting = nm.connection._ConnectionSetting(mock.MagicMock())
    con_setting.set_controller("controller0", "slave-type")

    assert con_setting.setting.props.master == "controller0"
    assert con_setting.setting.props.slave_type == "slave-type"



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.ifaces import Ifaces
from libnmstate.ifaces.bond import BondIface

from ..testlib.constants import FOO_IFACE_NAME
from ..testlib.constants import MAC_ADDRESS1
from ..testlib.ifacelib import gen_foo_iface_info

PORT1_IFACE_NAME = "port1"
PORT2_IFACE_NAME = "port2"


TEST_PORT = [PORT1_IFACE_NAME, PORT2_IFACE_NAME]
TEST_BOND_MODE = BondMode.ROUND_ROBIN

parametrize_bond_named_options = pytest.mark.parametrize(
    ("option_name", "name_value", "int_value"),
    (
        ("ad_select", "stable", 0),
        ("ad_select", "bandwidth", 1),
        ("ad_select", "count", 2),
        ("arp_validate", "active", 1),
        ("arp_validate", "filter", 4),
        ("arp_validate", "filter_backup", 6),
        ("fail_over_mac", "none", 0),
        ("fail_over_mac", "active", 1),
        ("fail_over_mac", "follow", 2),
    ),
)


class TestBondIface:
    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.BOND)
        iface_info[Bond.CONFIG_SUBTREE] = {
            Bond.MODE: TEST_BOND_MODE,
            Bond.PORT: deepcopy(TEST_PORT),
            Bond.OPTIONS_SUBTREE: {},
        }
        return iface_info

    def _gen_port1_iface_info(self):
        iface_info = gen_foo_iface_info()
        iface_info[Interface.NAME] = PORT1_IFACE_NAME
        return iface_info

    def _gen_port2_iface_info(self):
        iface_info = gen_foo_iface_info()
        iface_info[Interface.NAME] = PORT2_IFACE_NAME
        return iface_info

    def _gen_ifaces(self, bond_iface_info):
        return Ifaces(
            des_iface_infos=[
                bond_iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
            cur_iface_infos=[],
        )

    def test_bond_sort_port(self):
        iface_info1 = self._gen_iface_info()
        iface_info2 = self._gen_iface_info()
        iface_info2[Bond.CONFIG_SUBTREE][Bond.PORT].reverse()

        assert iface_info2 != iface_info1
        assert (
            BondIface(iface_info1).state_for_verify()
            == BondIface(iface_info2).state_for_verify()
        )

    def test_normalize_bond_option_str_to_integer(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "miimon": "140"
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "miimon": 140
        }

        assert BondIface(iface_info).to_dict() == expected_iface_info

    @parametrize_bond_named_options
    def test_normalize_bond_option_int_to_named(
        self, option_name, name_value, int_value
    ):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            option_name: int_value,
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            option_name: name_value
        }

        assert BondIface(iface_info).to_dict() == expected_iface_info

    @parametrize_bond_named_options
    def test_normalize_bond_option_int_str_to_named(
        self, option_name, name_value, int_value
    ):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            option_name: f"{int_value}",
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            option_name: name_value
        }

        assert BondIface(iface_info).to_dict() == expected_iface_info

    def test_normalize_bond_named_option_invalid_choice_preserved_as_int(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "ad_select": "99",
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "ad_select": 99
        }

        assert BondIface(iface_info).to_dict() == expected_iface_info

    def test_include_arp_ip_target_explictly_when_disable(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "arp_interval": 0,
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "arp_interval": 0,
            "arp_ip_target": "",
        }

        assert BondIface(iface_info).to_dict() == expected_iface_info

    def test_include_arp_ip_target_explictly_when_disable_in_verify(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "arp_interval": 0,
        }
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "arp_interval": 0,
            "arp_ip_target": "",
        }

        assert BondIface(iface_info).state_for_verify() == expected_iface_info

    def test_get_port(self):
        assert BondIface(self._gen_iface_info()).port == TEST_PORT

    def test_is_controller(self):
        assert BondIface(self._gen_iface_info()).is_controller

    def test_is_virtual(self):
        assert BondIface(self._gen_iface_info()).is_virtual

    def test_get_bond_mode(self):
        assert BondIface(self._gen_iface_info()).bond_mode == TEST_BOND_MODE

    def test_get_is_bond_mode_changed(self):
        cur_iface_info = self._gen_iface_info()
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP

        ifaces = Ifaces(
            des_iface_infos=[
                iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
            cur_iface_infos=[
                cur_iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
        )
        bond_iface = ifaces.all_kernel_ifaces[FOO_IFACE_NAME]
        bond_iface.gen_metadata(ifaces)
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]

        assert port1_iface.controller == FOO_IFACE_NAME
        assert port2_iface.controller == FOO_IFACE_NAME
        assert bond_iface.is_bond_mode_changed

    def test_pre_edit_clean_up_discard_bond_options_when_mode_chagned(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "miimon": 140
        }
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "arp_interval": 140
        }
        expected_iface_info = deepcopy(iface_info)

        ifaces = Ifaces(
            des_iface_infos=[
                iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
            cur_iface_infos=[
                cur_iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
        )
        bond_iface = ifaces.all_kernel_ifaces[FOO_IFACE_NAME]
        bond_iface.gen_metadata(ifaces)
        bond_iface.pre_edit_validation_and_cleanup()

        assert expected_iface_info == _remove_metadata(bond_iface.to_dict())

    def test_is_mac_restricted_mode(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "fail_over_mac": "active"
        }
        assert BondIface(iface_info).is_mac_restricted_mode

    def test_pre_edit_clean_up_discard_mac_address_when_mac_restricted(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Interface.MAC] = MAC_ADDRESS1
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "fail_over_mac": "active"
        }
        expected_iface_info = deepcopy(iface_info)

        ifaces = Ifaces(
            des_iface_infos=[
                iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
            cur_iface_infos=[
                cur_iface_info,
                self._gen_port1_iface_info(),
                self._gen_port2_iface_info(),
            ],
        )
        bond_iface = ifaces.all_kernel_ifaces[FOO_IFACE_NAME]
        bond_iface.gen_metadata(ifaces)
        bond_iface.pre_edit_validation_and_cleanup()

        assert expected_iface_info == _remove_metadata(bond_iface.to_dict())

    def test_validate_bond_mode_undefined(self):
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE].pop(Bond.MODE)

        with pytest.raises(NmstateValueError):
            BondIface(iface_info).pre_edit_validation_and_cleanup()

    def test_validate_mac_restriced_mode_with_desire_has_no_mac(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Interface.MAC] = MAC_ADDRESS1
        cur_iface = BondIface(cur_iface_info)
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "fail_over_mac": "active"
        }
        iface = BondIface(iface_info)

        iface.merge(cur_iface)
        iface.pre_edit_validation_and_cleanup()

    def test_validate_mac_restriced_mode_with_desire_has_mac(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Interface.MAC] = MAC_ADDRESS1
        cur_iface = BondIface(cur_iface_info)
        iface_info = self._gen_iface_info()
        iface_info[Interface.MAC] = MAC_ADDRESS1
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "fail_over_mac": "active"
        }
        iface = BondIface(iface_info)

        iface.merge(cur_iface)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_mac_restriced_mode_with_desire_has_mac_only(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP
        cur_iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE] = {
            "fail_over_mac": "active"
        }
        cur_iface = BondIface(cur_iface_info)
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE].pop(Bond.MODE)
        iface_info[Bond.CONFIG_SUBTREE].pop(Bond.OPTIONS_SUBTREE)
        iface_info[Interface.MAC] = MAC_ADDRESS1
        iface = BondIface(iface_info)

        iface.merge(cur_iface)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_remove_port(self):
        iface = BondIface(self._gen_iface_info())
        expected_iface_info = self._gen_iface_info()
        expected_iface_info[Bond.CONFIG_SUBTREE][Bond.PORT] = [
            PORT2_IFACE_NAME
        ]

        iface.remove_port(PORT1_IFACE_NAME)

        assert iface.to_dict() == expected_iface_info

    def test_match_ignore_bond_options(self):
        cur_iface_info = self._gen_iface_info()
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE]["foo"] = "bar"

        iface = BondIface(iface_info)
        cur_iface = BondIface(cur_iface_info)

        assert not iface.match(cur_iface)
        assert iface.match_ignore_bond_options(cur_iface)

    def test_match_ignore_bond_options_with_mode_different_only(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ROUND_ROBIN
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.MODE] = BondMode.ACTIVE_BACKUP

        iface = BondIface(iface_info)
        cur_iface = BondIface(cur_iface_info)

        assert not iface.match(cur_iface)
        assert not iface.match_ignore_bond_options(cur_iface)

    def test_match_ignore_bond_options_with_port_different_only(self):
        cur_iface_info = self._gen_iface_info()
        cur_iface_info[Bond.CONFIG_SUBTREE][Bond.PORT] = []
        iface_info = self._gen_iface_info()
        iface_info[Bond.CONFIG_SUBTREE][Bond.PORT] = ["foo"]

        iface = BondIface(iface_info)
        cur_iface = BondIface(cur_iface_info)

        assert not iface.match(cur_iface)
        assert not iface.match_ignore_bond_options(cur_iface)


def _remove_metadata(bond_iface_info):
    remove_keys = [k for k in bond_iface_info.keys() if k.startswith("_")]
    for key in remove_keys:
        bond_iface_info.pop(key)
    return bond_iface_info



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import json
import os

import pytest

import libnmstate
from libnmstate.error import NmstateDependencyError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Team
from libnmstate.error import NmstateValueError

from .testlib import assertlib
from .testlib import statelib
from .testlib.cmdlib import exec_cmd
from .testlib.cmdlib import RC_SUCCESS
from .testlib.nmplugin import disable_nm_plugin


TEAM0 = "team0"
PORT1 = "eth1"
PORT2 = "eth2"


pytestmark = pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="Team kmod not available in Travis CI",
)


def test_create_team_iface_without_port():
    with team_interface(TEAM0) as team_state:
        assertlib.assert_state(team_state)


@pytest.mark.tier1
def test_create_team_iface_with_port(eth1_up, eth2_up):
    with team_interface(TEAM0, [PORT1, PORT2]) as team_state:
        assertlib.assert_state_match(team_state)
        assert [PORT1, PORT2] == _get_runtime_team_port(TEAM0)
    assertlib.assert_absent(TEAM0)


def test_edit_team_iface(eth1_up):
    with team_interface(TEAM0) as team_state:
        team_state[Interface.KEY][0][Team.CONFIG_SUBTREE] = {
            Team.RUNNER_SUBTREE: {
                Team.Runner.NAME: Team.Runner.RunnerMode.LOAD_BALANCE
            },
            Team.PORT_SUBTREE: [{Team.Port.NAME: "eth1"}],
        }
        libnmstate.apply(team_state)
        assertlib.assert_state_match(team_state)


def test_nm_team_plugin_missing():
    with disable_nm_plugin("team"):
        with pytest.raises(NmstateDependencyError):
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: TEAM0,
                            Interface.TYPE: InterfaceType.TEAM,
                            Interface.STATE: InterfaceState.UP,
                        }
                    ]
                }
            )


def test_add_invalid_port_ip_config(eth1_up):
    desired_state = eth1_up
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.ENABLED] = True
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.DHCP] = True
    with pytest.raises(NmstateValueError):
        with team_interface(TEAM0, ports=("eth1",)) as state:
            desired_state[Interface.KEY].append(state[Interface.KEY][0])
            libnmstate.apply(desired_state)


@contextmanager
def team_interface(ifname, ports=None):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: ifname,
                Interface.TYPE: InterfaceType.TEAM,
                Interface.STATE: InterfaceState.UP,
            }
        ]
    }
    if ports:
        team_state = {Team.PORT_SUBTREE: []}
        desired_state[Interface.KEY][0][Team.CONFIG_SUBTREE] = team_state
        for port in ports:
            team_state[Team.PORT_SUBTREE].append({Team.Port.NAME: port})
    libnmstate.apply(desired_state)
    try:
        yield desired_state
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: ifname,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )


def _get_runtime_team_port(team_iface_name):
    """
    Use `teamdctl team0 state dump` to check team runtime status
    """
    rc, output, _ = exec_cmd(f"teamdctl {team_iface_name} state dump".split())
    assert rc == RC_SUCCESS
    teamd_state = json.loads(output)
    return sorted(teamd_state.get("ports", {}).keys())


def test_team_change_port_order(eth1_up, eth2_up):
    with team_interface(TEAM0, [PORT1, PORT2]) as desired_state:
        desired_state[Interface.KEY][0][Team.CONFIG_SUBTREE][
            Team.PORT_SUBTREE
        ].reverse()
        libnmstate.apply(desired_state)


@pytest.fixture
def team0_up(eth1_up, eth2_up):
    with team_interface(TEAM0, [PORT1, PORT2]):
        yield


def test_show_saved_config_with_team_down(team0_up):
    running_state = statelib.show_only((TEAM0,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEAM0,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((TEAM0,))

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import pytest

from unittest import mock

from libnmstate import netinfo
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Route
from libnmstate.schema import RouteRule


@pytest.fixture
def show_with_plugins_mock():
    with mock.patch.object(netinfo, "show_with_plugins") as m:
        yield m


@pytest.fixture
def plugin_context_mock():
    with mock.patch.object(netinfo, "plugin_context") as m:

        def enter(self):
            return self

        m().__enter__ = enter
        yield m


def test_netinfo_show(show_with_plugins_mock, plugin_context_mock):
    current_config = {
        DNS.KEY: {DNS.RUNNING: {}, DNS.CONFIG: {}},
        Route.KEY: {Route.CONFIG: [], Route.RUNNING: []},
        RouteRule.KEY: {RouteRule.CONFIG: []},
        Interface.KEY: [
            {
                Interface.NAME: "foo",
                Interface.TYPE: InterfaceType.UNKNOWN,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ],
    }

    show_with_plugins_mock.return_value = current_config
    report = netinfo.show()

    assert current_config == report


def test_error_show(show_with_plugins_mock, plugin_context_mock):
    current_config = {
        DNS.KEY: {DNS.RUNNING: {}, DNS.CONFIG: {}},
        Route.KEY: {"config": [], "running": []},
        RouteRule.KEY: {RouteRule.CONFIG: []},
        Interface.KEY: [
            {
                "name": "foo",
                "type": "unknown",
                "state": "up",
                "ipv4": {InterfaceIPv4.ENABLED: False},
                "ipv6": {InterfaceIPv6.ENABLED: False},
            }
        ],
    }
    show_with_plugins_mock.return_value = current_config

    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        netinfo.show(None)
        # pylint: enable=too-many-function-args



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from nispor import NisporNetState

from libnmstate.plugin import NmstatePlugin
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .base_iface import NisporPluginBaseIface
from .bond import NisporPluginBondIface
from .bridge import NisporPluginBridgeIface
from .dummy import NisporPluginDummyIface
from .ethernet import NisporPluginEthernetIface
from .macvlan import NisporPluginMacVlanIface
from .macvtap import NisporPluginMacVtapIface
from .vlan import NisporPluginVlanIface
from .vxlan import NisporPluginVxlanIface
from .route import nispor_route_state_to_nmstate
from .route_rule import nispor_route_rule_state_to_nmstate
from .vrf import NisporPluginVrfIface
from .ovs import NisporPluginOvsInternalIface


_INFO_TYPE_RUNNING = "running"
_INFO_TYPE_RUNNING_CONFIG = "running_config"


class NisporPlugin(NmstatePlugin):
    @property
    def name(self):
        return "nispor"

    @property
    def plugin_capabilities(self):
        return [
            NmstatePlugin.PLUGIN_CAPABILITY_IFACE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE_RULE,
        ]

    @property
    def priority(self):
        # Let NetworkManagerPlugin take priority as nispor might be wrong on
        # DHCP status and currently nispor does not support all interface types
        # yet.
        return NmstatePlugin.DEFAULT_PRIORITY - 1

    def _get_interfaces(self, info_type):
        np_state = NisporNetState.retrieve()
        ifaces = []
        config_only = info_type == _INFO_TYPE_RUNNING_CONFIG
        for np_iface in np_state.ifaces.values():
            iface_type = np_iface.type
            if iface_type == "dummy":
                ifaces.append(
                    NisporPluginDummyIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "veth":
                ifaces.append(
                    NisporPluginEthernetIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "ethernet":
                ifaces.append(
                    NisporPluginEthernetIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "bond":
                ifaces.append(
                    NisporPluginBondIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "vlan":
                ifaces.append(
                    NisporPluginVlanIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "vxlan":
                ifaces.append(
                    NisporPluginVxlanIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "mac_vlan":
                ifaces.append(
                    NisporPluginMacVlanIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "mac_vtap":
                ifaces.append(
                    NisporPluginMacVtapIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "bridge":
                np_ports = []
                for port_name in np_iface.ports:
                    if port_name in np_state.ifaces.keys():
                        np_ports.append(np_state.ifaces[port_name])
                ifaces.append(
                    NisporPluginBridgeIface(np_iface, np_ports).to_dict(
                        config_only
                    )
                )
            elif iface_type == "vrf":
                ifaces.append(
                    NisporPluginVrfIface(np_iface).to_dict(config_only)
                )
            elif iface_type == "openv_switch":
                ifaces.append(
                    NisporPluginOvsInternalIface(np_iface).to_dict(config_only)
                )
            else:
                ifaces.append(
                    NisporPluginBaseIface(np_iface).to_dict(config_only)
                )
        return ifaces

    def get_interfaces(self):
        return self._get_interfaces(_INFO_TYPE_RUNNING)

    def get_running_config_interfaces(self):
        return self._get_interfaces(_INFO_TYPE_RUNNING_CONFIG)

    def get_routes(self):
        np_state = NisporNetState.retrieve()
        return {Route.RUNNING: nispor_route_state_to_nmstate(np_state.routes)}

    def get_route_rules(self):
        np_state = NisporNetState.retrieve()
        return {
            RouteRule.CONFIG: nispor_route_rule_state_to_nmstate(
                np_state.route_rules
            )
        }



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections import defaultdict
from distutils.version import StrictVersion
import logging
from operator import itemgetter

from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateValueError
from libnmstate.ifaces.base_iface import BaseIface
from libnmstate.ifaces.ovs import is_ovs_running
from libnmstate.plugin import NmstatePlugin
from libnmstate.schema import Bond
from libnmstate.schema import DNS
from libnmstate.schema import Ethernet
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import LLDP
from libnmstate.schema import LinuxBridge
from libnmstate.schema import MacVlan
from libnmstate.schema import MacVtap
from libnmstate.schema import OVSBridge
from libnmstate.schema import OVSInterface
from libnmstate.schema import Route
from libnmstate.schema import Team
from libnmstate.schema import VLAN
from libnmstate.schema import VRF
from libnmstate.schema import VXLAN


from .bond import get_bond_config
from .bridge import get_linux_bridge_config
from .checkpoint import CheckPoint
from .checkpoint import get_checkpoints
from .common import NM
from .context import NmContext
from .device import get_device_common_info
from .device import list_devices
from .dns import get_running as get_dns_running
from .dns import get_dns_config_from_nm_profiles
from .infiniband import get_infiniband_config
from .ipv4 import get_ipv4_config
from .ipv6 import get_ipv6_config
from .lldp import get_lldp_config
from .macvlan import get_current_macvlan_type
from .macvlan import get_macvtap_config
from .ovs import get_interface_info as get_ovs_interface_info
from .ovs import get_ovs_bridge_config
from .ovs import get_ovs_bridge_info
from .ovs import get_ovs_patch_iface_config
from .ovs import has_ovs_capability
from .profile import get_basic_iface_info
from .profiles import NmProfiles
from .profiles import get_all_applied_configs
from .route import get_routes_from_nm_profiles
from .route import get_route_rules_from_nm_profiles
from .sriov import get_sriov_config
from .team import get_info as get_team_info
from .team import get_team_config
from .team import has_team_capability
from .translator import Nm2Api
from .user import get_user_config
from .vlan import get_vlan_config
from .vrf import get_vrf_config
from .vxlan import get_vxlan_config
from .wired import get_info as get_wired_info
from .wired import get_wired_config

_METADATA_PROFILE_PRIORITY = "_priority"
_METADATA_PROFILE_TIMESTAMP = "_timestamp"
_METADATA_NM_PROFILE = "_nm_profile"


class NetworkManagerPlugin(NmstatePlugin):
    def __init__(self):
        self._ctx = NmContext()
        self._checkpoint = None
        self._check_version_mismatch()
        self.__applied_configs = None

    @property
    def priority(self):
        return NmstatePlugin.DEFAULT_PRIORITY

    @property
    def name(self):
        return "NetworkManager"

    def unload(self):
        if self._ctx:
            self._ctx.clean_up()
            self._ctx = None

    @property
    def _applied_configs(self):
        if self.__applied_configs is None:
            self.__applied_configs = get_all_applied_configs(self.context)
        return self.__applied_configs

    @property
    def checkpoint(self):
        return self._checkpoint

    @property
    def client(self):
        return self._ctx.client if self._ctx else None

    @property
    def context(self):
        return self._ctx

    @property
    def capabilities(self):
        capabilities = []
        if has_ovs_capability(self.client) and is_ovs_running():
            capabilities.append(NmstatePlugin.OVS_CAPABILITY)
        if has_team_capability(self.client):
            capabilities.append(NmstatePlugin.TEAM_CAPABILITY)
        return capabilities

    @property
    def plugin_capabilities(self):
        return [
            NmstatePlugin.PLUGIN_CAPABILITY_IFACE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE_RULE,
            NmstatePlugin.PLUGIN_CAPABILITY_DNS,
        ]

    def get_interfaces(self):
        info = []

        applied_configs = self._applied_configs

        devices_info = [
            (dev, get_device_common_info(dev))
            for dev in list_devices(self.client)
        ]

        for dev, devinfo in devices_info:
            if not dev.get_managed():
                # Skip unmanaged interface
                continue

            iface_info = Nm2Api.get_common_device_info(devinfo)
            applied_config = applied_configs.get(iface_info[Interface.NAME])

            act_con = dev.get_active_connection()
            iface_info[Interface.IPV4] = get_ipv4_config(applied_config)
            iface_info[Interface.IPV6] = get_ipv6_config(applied_config)
            iface_info.update(get_wired_info(dev))
            iface_info.update(get_user_config(applied_config))
            iface_info[LLDP.CONFIG_SUBTREE] = get_lldp_config(
                applied_config, nmdev=dev
            )
            iface_info.update(get_team_info(dev))
            iface_info.update(get_infiniband_config(applied_config))
            iface_info.update(get_current_macvlan_type(applied_config))

            if iface_info[Interface.TYPE] == InterfaceType.OVS_BRIDGE:
                iface_info.update(get_ovs_bridge_info(dev))
                iface_info = _remove_ovs_bridge_unsupported_entries(iface_info)
            elif iface_info[Interface.TYPE] == InterfaceType.OVS_INTERFACE:
                iface_info.update(get_ovs_interface_info(act_con))
            elif iface_info[Interface.TYPE] == InterfaceType.OVS_PORT:
                continue

            info.append(iface_info)

        info.sort(key=itemgetter("name"))

        return info

    def get_running_config_interfaces(self):
        iface_infos = self.get_interfaces()
        # Remove LLDP neighber information
        for iface_info in iface_infos:
            if LLDP.CONFIG_SUBTREE in iface_info:
                iface_info[LLDP.CONFIG_SUBTREE].pop(
                    LLDP.NEIGHBORS_SUBTREE, None
                )
        return iface_infos

    def _get_iface_info_from_saved_profiles(self):
        iface_infos = {}
        for nm_profile in self.client.get_connections():
            iface_info = get_basic_iface_info(nm_profile)
            iface_type = iface_info[Interface.TYPE]
            iface_name = iface_info[Interface.NAME]
            nm_conn_setting = nm_profile.get_setting_connection()
            if not nm_conn_setting:
                continue
            iface_info[
                _METADATA_PROFILE_PRIORITY
            ] = nm_conn_setting.get_autoconnect_priority()
            iface_info[
                _METADATA_PROFILE_TIMESTAMP
            ] = nm_conn_setting.get_timestamp()
            if not iface_name or iface_type == InterfaceType.UNKNOWN:
                # TODO: Support matching interface via MAC
                continue
            iface_index = iface_info[Interface.NAME]
            if iface_type in (
                InterfaceType.OVS_BRIDGE,
                InterfaceType.OVS_PORT,
            ):
                iface_index = f"{iface_type} {iface_name}"
            cur_iface_info = iface_infos.get(iface_index)
            should_save = False
            if not cur_iface_info:
                should_save = True
            else:
                cur_priority = cur_iface_info[_METADATA_PROFILE_PRIORITY]
                new_priority = iface_infos[_METADATA_PROFILE_PRIORITY]
                if new_priority > cur_priority:
                    should_save = True
                elif new_priority == cur_priority:
                    cur_timestamp = cur_iface_info[_METADATA_PROFILE_TIMESTAMP]
                    new_timestamp = iface_info[_METADATA_PROFILE_TIMESTAMP]
                    if new_timestamp > cur_timestamp:
                        should_save = True

            if should_save:
                iface_info[_METADATA_NM_PROFILE] = nm_profile
                iface_name = iface_info[Interface.NAME]
                iface_type = iface_info[Interface.TYPE]
                if iface_type in (
                    InterfaceType.OVS_BRIDGE,
                    InterfaceType.OVS_PORT,
                ):
                    iface_infos[f"{iface_type} {iface_name}"] = iface_info
                else:
                    iface_infos[iface_name] = iface_info
        return iface_infos

    def get_saved_config_interfaces(self):
        iface_infos = self._get_iface_info_from_saved_profiles()

        _fill_interface_specific_info(iface_infos)

        for iface_info in iface_infos.values():
            metadata_keys = [
                key for key in iface_info.keys() if key.startswith("_")
            ]
            for key in metadata_keys:
                iface_info.pop(key)

        return sorted(
            [
                iface_info
                for iface_info in iface_infos.values()
                if iface_info[Interface.TYPE] != InterfaceType.OVS_PORT
            ],
            key=itemgetter(Interface.NAME),
        )

    def get_routes(self):
        return {
            Route.CONFIG: get_routes_from_nm_profiles(self._applied_configs)
        }

    def get_saved_routes(self):
        iface_infos = self._get_iface_info_from_saved_profiles()
        return get_routes_from_nm_profiles(
            {
                iface_info[Interface.NAME]: iface_info[_METADATA_NM_PROFILE]
                for iface_info in iface_infos.values()
            }
        )

    def get_route_rules(self):
        """
        Nispor will provide running config of route rule from kernel.
        """
        return {}

    def get_saved_route_rules(self):
        iface_infos = self._get_iface_info_from_saved_profiles()
        return get_route_rules_from_nm_profiles(
            [
                iface_info[_METADATA_NM_PROFILE]
                for iface_info in iface_infos.values()
            ]
        )

    def get_dns_client_config(self):
        return {
            DNS.RUNNING: get_dns_running(self.client),
            DNS.CONFIG: get_dns_config_from_nm_profiles(self._applied_configs),
        }

    def get_saved_dns_client_config(self):
        iface_infos = self._get_iface_info_from_saved_profiles()
        return get_dns_config_from_nm_profiles(
            {
                iface_index: iface_info[_METADATA_NM_PROFILE]
                for iface_index, iface_info in iface_infos.items()
            }
        )

    def refresh_content(self):
        self.__applied_configs = None
        self._ctx.refresh_content()

    def apply_changes(self, net_state, save_to_disk):
        NmProfiles(self.context).apply_config(net_state, save_to_disk)

    def _load_checkpoint(self, checkpoint_path):
        if checkpoint_path:
            if self._checkpoint:
                # Old checkpoint might timeout, hence it's legal to load
                # another one.
                self._checkpoint.clean_up()
            candidates = get_checkpoints(self._ctx.client)
            if checkpoint_path in candidates:
                self._checkpoint = CheckPoint(
                    nm_context=self._ctx, dbuspath=checkpoint_path
                )
            else:
                raise NmstateValueError("No checkpoint specified or found")
        else:
            if not self._checkpoint:
                # Get latest one
                candidates = get_checkpoints(self._ctx.client)
                if candidates:
                    self._checkpoint = CheckPoint(
                        nm_context=self._ctx, dbuspath=candidates[0]
                    )
                else:
                    raise NmstateValueError("No checkpoint specified or found")

    def create_checkpoint(self, timeout=60):
        self._checkpoint = CheckPoint.create(self._ctx, timeout)
        return str(self._checkpoint)

    def rollback_checkpoint(self, checkpoint=None):
        self._load_checkpoint(checkpoint)
        self._checkpoint.rollback()
        self._checkpoint = None

    def destroy_checkpoint(self, checkpoint=None):
        self._load_checkpoint(checkpoint)
        self._checkpoint.destroy()
        self._checkpoint = None

    def _check_version_mismatch(self):
        nm_client_version = self._ctx.client.get_version()
        nm_utils_version = _nm_utils_decode_version()

        if nm_client_version is None:
            raise NmstateDependencyError(
                "NetworkManager daemon is not running which is required for "
                "NetworkManager plugin"
            )
        elif StrictVersion(nm_client_version) != StrictVersion(
            nm_utils_version
        ):
            logging.warning(
                "libnm version %s mismatches NetworkManager version %s",
                nm_utils_version,
                nm_client_version,
            )


def _remove_ovs_bridge_unsupported_entries(iface_info):
    """
    OVS bridges are not supporting several common interface key entries.
    These entries are removed explicitly.
    """
    iface_info.pop(Interface.IPV4, None)
    iface_info.pop(Interface.IPV6, None)
    iface_info.pop(Interface.MTU, None)

    return iface_info


def _nm_utils_decode_version():
    return f"{NM.MAJOR_VERSION}.{NM.MINOR_VERSION}.{NM.MICRO_VERSION}"


def _fill_interface_specific_info(iface_infos):
    for iface_index, iface_info in iface_infos.items():
        iface_name = iface_info[Interface.NAME]
        nm_profile = iface_info[_METADATA_NM_PROFILE]
        subordinate_nm_profiles = [
            tmp_iface_info[_METADATA_NM_PROFILE]
            for tmp_iface_info in iface_infos.values()
            if tmp_iface_info[BaseIface.CONTROLLER_METADATA] == iface_name
            and tmp_iface_info[BaseIface.CONTROLLER_TYPE_METADATA]
            == nm_profile.get_connection_type()
        ]
        iface_info[Interface.IPV4] = get_ipv4_config(
            nm_profile, full_config=True
        )
        iface_info[Interface.IPV6] = get_ipv6_config(
            nm_profile, full_config=True
        )
        iface_info.update(get_wired_config(nm_profile))
        iface_info.update(get_user_config(nm_profile))
        iface_info[LLDP.CONFIG_SUBTREE] = get_lldp_config(
            nm_profile, config_only=True
        )
        sriov_config = get_sriov_config(nm_profile)
        if sriov_config:
            iface_info[Ethernet.SRIOV_SUBTREE] = sriov_config

        iface_type = iface_info[Interface.TYPE]
        if iface_type == InterfaceType.TEAM:
            iface_info[Team.CONFIG_SUBTREE] = get_team_config(
                nm_profile, subordinate_nm_profiles
            )
        elif iface_type == InterfaceType.INFINIBAND:
            iface_info.update(get_infiniband_config(nm_profile))
        elif iface_type == InterfaceType.OVS_BRIDGE:
            _fill_ovs_bridge_iface_info(
                iface_info, iface_infos, subordinate_nm_profiles
            )
        elif iface_type == InterfaceType.OVS_INTERFACE:
            ovs_patch_info = get_ovs_patch_iface_config(nm_profile)
            if ovs_patch_info:
                iface_info[OVSInterface.PATCH_CONFIG_SUBTREE] = ovs_patch_info
        elif iface_type == InterfaceType.BOND:
            iface_info[Bond.CONFIG_SUBTREE] = get_bond_config(
                nm_profile, subordinate_nm_profiles
            )
        elif iface_type == InterfaceType.VLAN:
            iface_info[VLAN.CONFIG_SUBTREE] = get_vlan_config(nm_profile)
        elif iface_type == InterfaceType.VXLAN:
            iface_info[VXLAN.CONFIG_SUBTREE] = get_vxlan_config(nm_profile)
        elif iface_type == InterfaceType.LINUX_BRIDGE:
            iface_info[LinuxBridge.CONFIG_SUBTREE] = get_linux_bridge_config(
                nm_profile, subordinate_nm_profiles
            )
        elif iface_type == InterfaceType.MAC_VTAP:
            iface_info[MacVtap.CONFIG_SUBTREE] = get_macvtap_config(nm_profile)
        elif iface_type == InterfaceType.MAC_VLAN:
            iface_info[MacVlan.CONFIG_SUBTREE] = get_macvtap_config(nm_profile)
        elif iface_type == InterfaceType.VRF:
            iface_info[VRF.CONFIG_SUBTREE] = get_vrf_config(
                nm_profile, subordinate_nm_profiles
            )


def _fill_ovs_bridge_iface_info(iface_info, iface_infos, ovs_port_nm_profiles):
    nm_profile = iface_info[_METADATA_NM_PROFILE]
    ovs_iface_nm_profiles = defaultdict(list)
    for nm_ovs_port_profile in ovs_port_nm_profiles:
        ovs_port_name = nm_ovs_port_profile.get_interface_name()
        for tmp_iface_info in iface_infos.values():
            if (
                tmp_iface_info[BaseIface.CONTROLLER_METADATA] == ovs_port_name
                and tmp_iface_info[BaseIface.CONTROLLER_TYPE_METADATA]
                == InterfaceType.OVS_PORT
            ):
                ovs_iface_nm_profiles[ovs_port_name].append(
                    tmp_iface_info[_METADATA_NM_PROFILE]
                )

    iface_info[OVSBridge.CONFIG_SUBTREE] = get_ovs_bridge_config(
        nm_profile, ovs_port_nm_profiles, ovs_iface_nm_profiles
    )
    _remove_ovs_bridge_unsupported_entries(iface_info)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import MacVlan
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.macvlan import MacVlanIface

from ..testlib.ifacelib import gen_foo_iface_info


BASE_IFACE_NAME = "base1"


class TestMacVlanIface:
    def _gen_base_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.ETHERNET)
        iface_info[Interface.NAME] = BASE_IFACE_NAME
        iface_info[Interface.MTU] = 1500

        return iface_info

    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.MAC_VLAN)
        iface_info[MacVlan.CONFIG_SUBTREE] = {
            MacVlan.BASE_IFACE: BASE_IFACE_NAME,
            MacVlan.MODE: MacVlan.Mode.PASSTHRU,
        }
        return iface_info

    def test_get_parent(self):
        assert MacVlanIface(self._gen_iface_info()).parent == BASE_IFACE_NAME

    def test_need_parent(self):
        assert MacVlanIface(self._gen_iface_info()).need_parent

    def test_is_virtual(self):
        assert MacVlanIface(self._gen_iface_info()).is_virtual

    def test_can_have_ip_when_enslaved(self):
        assert not MacVlanIface(
            self._gen_iface_info()
        ).can_have_ip_when_enslaved

    def test_add_macvlan_without_mode(self):
        iface_info = self._gen_iface_info()
        iface_info[MacVlan.CONFIG_SUBTREE].pop(MacVlan.MODE)

        iface = MacVlanIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_add_macvlan_without_base_iface(self):
        iface_info = self._gen_iface_info()
        iface_info[MacVlan.CONFIG_SUBTREE].pop(MacVlan.BASE_IFACE)

        iface = MacVlanIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_add_macvlan_not_passthru_and_not_promisc(self):
        iface_info = self._gen_iface_info()
        iface_info[MacVlan.CONFIG_SUBTREE][MacVlan.MODE] = MacVlan.Mode.VEPA
        iface_info[MacVlan.CONFIG_SUBTREE][MacVlan.PROMISCUOUS] = False

        iface = MacVlanIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from unittest import mock

from libnmstate import nm


@pytest.fixture
def NM_mock():
    with mock.patch.object(nm.user, "NM") as m:
        yield m


def test_create_no_setting(NM_mock):
    setting = nm.user.create_setting({}, None)
    assert setting is None


def test_create_setting_duplicate(NM_mock):
    base_profile = mock.MagicMock()

    setting = nm.user.create_setting(
        {"description": "test_interface"}, base_profile
    )
    base_profile.get_setting_by_name.assert_called_with(
        NM_mock.SETTING_USER_SETTING_NAME
    )
    assert (
        setting
        == base_profile.get_setting_by_name.return_value.duplicate.return_value
    )


def test_create_setting_description(NM_mock):
    setting = nm.user.create_setting({"description": "test_interface"}, None)
    assert setting == NM_mock.SettingUser.new.return_value
    setting.set_data.assert_called_with(
        "nmstate.interface.description", "test_interface"
    )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.ifaces import Ifaces
from libnmstate.schema import InfiniBand
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge
from libnmstate.schema import OVSBridge

from libnmstate.ifaces.infiniband import InfiniBandIface

from ..testlib.ifacelib import gen_foo_iface_info
from ..testlib.bridgelib import gen_bridge_iface_info
from ..testlib.ovslib import gen_ovs_bridge_info

BASE_IFACE_NAME = "ib0"
TEST_PKEY = "0x8001"
PKEY_IFACE_NAME = f"{BASE_IFACE_NAME}.{TEST_PKEY[2:]}"


class TestInfiniBandIface:
    def _gen_base_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.INFINIBAND)
        iface_info[Interface.NAME] = BASE_IFACE_NAME
        iface_info[Interface.MTU] = 1500
        iface_info[InfiniBand.CONFIG_SUBTREE] = {
            InfiniBand.PKEY: InfiniBand.DEFAULT_PKEY,
            InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
        }

        return iface_info

    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.INFINIBAND)
        iface_info[Interface.NAME] = PKEY_IFACE_NAME
        iface_info[InfiniBand.CONFIG_SUBTREE] = {
            InfiniBand.PKEY: TEST_PKEY,
            InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
            InfiniBand.BASE_IFACE: BASE_IFACE_NAME,
        }
        return iface_info

    def test_get_parent(self):
        assert (
            InfiniBandIface(self._gen_iface_info()).parent == BASE_IFACE_NAME
        )

    def test_need_parent(self):
        assert InfiniBandIface(self._gen_iface_info()).need_parent

    def test_base_iface_not_need_parent(self):
        assert not InfiniBandIface(self._gen_base_iface_info()).need_parent

    def test_is_not_virtual(self):
        assert not InfiniBandIface(self._gen_iface_info()).is_virtual

    def test_cannot_have_ip_as_port(self):
        assert not InfiniBandIface(self._gen_iface_info()).can_have_ip_as_port

    def test_validate_base_iface_missing(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE].pop(InfiniBand.BASE_IFACE)

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_mode_missing(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE].pop(InfiniBand.MODE)

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_cannonicalize_undefined_pkey(self):
        iface_info = self._gen_base_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE].pop(InfiniBand.PKEY)

        iface = InfiniBandIface(iface_info)
        iface.pre_edit_validation_and_cleanup()
        assert (
            iface.to_dict()[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY]
            == InfiniBand.DEFAULT_PKEY
        )

    def test_cannonicalize_hex_string_pkey_to_numer(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY] = TEST_PKEY

        iface = InfiniBandIface(iface_info)
        iface.pre_edit_validation_and_cleanup()
        assert iface.to_dict()[InfiniBand.CONFIG_SUBTREE][
            InfiniBand.PKEY
        ] == int(TEST_PKEY, 16)

    def test_cannonicalize_string_pkey_to_numer(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][
            InfiniBand.PKEY
        ] = f"{int(TEST_PKEY, 16)}"

        iface = InfiniBandIface(iface_info)
        iface.pre_edit_validation_and_cleanup()
        assert iface.to_dict()[InfiniBand.CONFIG_SUBTREE][
            InfiniBand.PKEY
        ] == int(TEST_PKEY, 16)

    def test_cannonicalize_invalid_string_pkey(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY] = "invalid_pkey"

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_cannonicalize_pkey_bigger_than_maximum(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY] = 0xFFFF + 1

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_cannonicalize_pkey_smaller_than_minimum(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY] = 0

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_invalid_iface_name(self):
        iface_info = self._gen_iface_info()
        iface_info[Interface.NAME] = "invalid_pkey"

        iface = InfiniBandIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_verify_string_pkey(self):
        iface_info = self._gen_iface_info()
        iface_info[InfiniBand.CONFIG_SUBTREE][InfiniBand.PKEY] = int(
            TEST_PKEY, 16
        )

        iface = InfiniBandIface(iface_info)
        current_iface = InfiniBandIface(self._gen_iface_info())
        iface.match(current_iface)

    def test_remove_base_iface_got_child_marked_as_absent(self):
        base_iface_info = self._gen_base_iface_info()
        base_iface_info[Interface.STATE] = InterfaceState.ABSENT
        ifaces = Ifaces([base_iface_info, self._gen_iface_info()], [])
        ifaces.all_kernel_ifaces[PKEY_IFACE_NAME].state = InterfaceState.ABSENT
        ifaces.all_kernel_ifaces[BASE_IFACE_NAME].state = InterfaceState.ABSENT

    def test_expect_exception_for_adding_ib_to_linux_bridge(self):
        base_iface_info = self._gen_base_iface_info()
        ib_iface_info = self._gen_iface_info()
        bridge_iface_info = gen_bridge_iface_info()
        bridge_config = bridge_iface_info[LinuxBridge.CONFIG_SUBTREE]
        bridge_config[LinuxBridge.PORT_SUBTREE] = [
            {LinuxBridge.Port.NAME: PKEY_IFACE_NAME}
        ]
        with pytest.raises(NmstateValueError):
            Ifaces([base_iface_info, ib_iface_info, bridge_iface_info], [])

    def test_expect_exception_for_adding_ib_to_ovs_bridge(self):
        base_iface_info = self._gen_base_iface_info()
        ib_iface_info = self._gen_iface_info()
        bridge_iface_info = gen_ovs_bridge_info()
        bridge_config = bridge_iface_info[OVSBridge.CONFIG_SUBTREE]
        bridge_config[OVSBridge.PORT_SUBTREE] = [
            {OVSBridge.Port.NAME: PKEY_IFACE_NAME}
        ]
        with pytest.raises(NmstateValueError):
            Ifaces([base_iface_info, ib_iface_info, bridge_iface_info], [])

    def test_add_ib_to_bond_not_in_active_backup_mode(self):
        base_iface_info = self._gen_base_iface_info()
        ib_iface_info = self._gen_iface_info()
        bond_iface_info = gen_foo_iface_info(iface_type=InterfaceType.BOND)
        bond_iface_info[Bond.CONFIG_SUBTREE] = {
            Bond.MODE: BondMode.ROUND_ROBIN,
            Bond.PORT: [PKEY_IFACE_NAME],
            Bond.OPTIONS_SUBTREE: {},
        }

        with pytest.raises(NmstateValueError):
            Ifaces([base_iface_info, ib_iface_info, bond_iface_info], [])

    def test_add_ib_to_bond_in_active_backup_mode(self):
        base_iface_info = self._gen_base_iface_info()
        ib_iface_info = self._gen_iface_info()
        bond_iface_info = gen_foo_iface_info(iface_type=InterfaceType.BOND)
        bond_iface_info[Bond.CONFIG_SUBTREE] = {
            Bond.MODE: BondMode.ACTIVE_BACKUP,
            Bond.PORT: [PKEY_IFACE_NAME],
            Bond.OPTIONS_SUBTREE: {},
        }

        Ifaces([base_iface_info, ib_iface_info, bond_iface_info], [])



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.state import state_match


class TestStateMatch:
    def test_match_empty_dict(self):
        assert state_match({}, {})

    def test_match_empty_list(self):
        assert state_match([], [])

    def test_match_none(self):
        assert state_match(None, None)

    def test_match_dict_vs_list(self):
        assert not state_match({}, [])

    def test_match_list_vs_string(self):
        assert not state_match(["a", "b", "c"], "abc")

    def test_match_dict_identical(self):
        assert state_match({"a": 1, "b": 2}, {"a": 1, "b": 2})

    def test_match_dict_current_has_more_data(self):
        assert state_match({"a": 1}, {"a": 1, "b": 2})

    def test_match_dict_desire_has_more_data(self):
        assert not state_match({"a": 1, "b": 2}, {"a": 1})

    def test_match_dict_different_value_type(self):
        assert not state_match({"a": 1, "b": []}, {"a": 1, "b": 2})

    def test_match_list_identical(self):
        assert state_match(["a", "b", 1], ["a", "b", 1])

    def test_match_list_different_order(self):
        assert not state_match(["a", "b", 1], ["a", 1, "b"])

    def test_match_list_current_contains_more(self):
        assert not state_match(["a", "b", 1], ["a", "b", "c", 1])

    def test_match_indentical_set(self):
        assert state_match(set(["a", "b", 1]), set(["a", "b", 1]))
        assert state_match(set(["a", 1, "b"]), set(["a", "b", 1]))
        assert state_match(set(["a", 1, 1, "b"]), set(["a", "b", 1]))

    def test_match_parital_set(self):
        assert not state_match(set(["a", "b", 1]), set(["a", "b", "c", 1]))

    def test_match_nested_list_in_dict(self):
        assert state_match({"a": 1, "b": [1, 2]}, {"a": 1, "b": [1, 2]})

    def test_match_nested_dict_in_list(self):
        assert state_match(
            [{"a": 1, "b": [1, 2]}, {"a": 2, "b": [3, 4]}],
            [{"a": 1, "b": [1, 2]}, {"a": 2, "b": [3, 4]}],
        )
        assert state_match(
            [{"a": 1}, {"a": 2, "b": [3, 4]}],
            [{"a": 1, "b": [1, 2]}, {"a": 2, "b": [3, 4]}],
        )
        assert not state_match(
            [{"a": 2, "b": [3, 4]}, {"a": 1, "b": [1, 2]}],
            [{"a": 1, "b": [1, 2]}, {"a": 2, "b": [3, 4]}],
        )



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import os

import yaml

import libnmstate


PATH_MAX = 4096


@contextmanager
def example_state(initial, cleanup=None, substitute=None):
    """
    Apply the initial state and optionally the cleanup state at the end
    """

    desired_state = load_example(initial, substitute)

    libnmstate.apply(desired_state)
    try:
        yield desired_state
    finally:
        if cleanup:
            try:
                libnmstate.apply(
                    load_example(cleanup, substitute), verify_change=True
                )
            except libnmstate.error.NmstateVerificationError:
                libnmstate.apply(
                    load_example(cleanup, substitute), verify_change=False
                )
                raise


def load_example(name, substitute=None):
    """
    Load the state from an example yaml file
    """

    examples = find_examples_dir()

    with open(os.path.join(examples, name)) as yamlfile:
        yaml_str = yamlfile.read()
        if substitute:
            yaml_str = yaml_str.replace(substitute[0], substitute[1])
        state = yaml.load(yaml_str, Loader=yaml.SafeLoader)

    return state


def find_examples_dir():
    """
    Look recursively for the directory containing the examples
    """

    path = ""
    parent = "../"
    rootdir = "/"
    examples = None
    for _ in range(PATH_MAX // len("x/")):
        maybe_examples = os.path.abspath(os.path.join(path, "examples"))
        if os.path.isdir(maybe_examples):
            examples = maybe_examples
            break

        if os.path.abspath(path) == rootdir:
            break

        path = parent + path

    if examples:
        return examples
    else:
        raise RuntimeError("Cannot find examples directory")



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Route
from libnmstate.route import RouteEntry

from .constants import IPV4_ADDRESS1
from .constants import IPV6_ADDRESS1

IPV4_ROUTE_IFACE_NAME = "ipv4_route_iface"
IPV4_ROUTE_DESITNATION = "198.51.100.0/24"
IPV4_DEFAULT_GATEWAY_DESTINATION = "0.0.0.0/0"
IPV4_ROUTE_TABLE_ID = 51
IPV6_ROUTE_IFACE_NAME = "ipv6_route_iface"
IPV6_ROUTE_DESITNATION = "2001:db8:a::/64"
IPV6_DEFAULT_GATEWAY_DESTINATION = "::/0"
IPV6_ROUTE_TABLE_ID = 52


def gen_ipv4_route():
    return _create_route(
        IPV4_ROUTE_DESITNATION,
        103,
        IPV4_ADDRESS1,
        IPV4_ROUTE_IFACE_NAME,
        IPV4_ROUTE_TABLE_ID,
    )


def gen_ipv6_route():
    return _create_route(
        IPV6_ROUTE_DESITNATION,
        104,
        IPV6_ADDRESS1,
        IPV6_ROUTE_IFACE_NAME,
        IPV6_ROUTE_TABLE_ID,
    )


def gen_ipv4_default_gateway():
    return _create_route(
        IPV4_DEFAULT_GATEWAY_DESTINATION,
        104,
        IPV4_ADDRESS1,
        IPV4_ROUTE_IFACE_NAME,
        52,
    )


def gen_ipv6_default_gateway():
    return _create_route(
        IPV6_DEFAULT_GATEWAY_DESTINATION,
        104,
        IPV6_ADDRESS1,
        IPV6_ROUTE_IFACE_NAME,
        52,
    )


def _create_route(dest, metric, via_addr, via_iface, table_id):
    return RouteEntry(
        {
            Route.DESTINATION: dest,
            Route.METRIC: metric,
            Route.NEXT_HOP_ADDRESS: via_addr,
            Route.NEXT_HOP_INTERFACE: via_iface,
            Route.TABLE_ID: table_id,
        }
    )



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from operator import itemgetter
import socket

from libnmstate import iplib
from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .common import GLib
from .common import NM

IPV4_DEFAULT_GATEWAY_DESTINATION = "0.0.0.0/0"
IPV6_DEFAULT_GATEWAY_DESTINATION = "::/0"

# NM require route rule priority been set explicitly, use 30,000 when
# desire state instruct to use USE_DEFAULT_PRIORITY
ROUTE_RULE_DEFAULT_PRIORIRY = 30000


def get_routes_from_nm_profiles(applied_configs):
    """
    Query routes saved in running profile
    """
    routes = []
    for iface_name, applied_config in applied_configs.items():
        for ip_profile in (
            applied_config.get_setting_ip6_config(),
            applied_config.get_setting_ip4_config(),
        ):
            if not ip_profile:
                continue
            nm_routes = ip_profile.props.routes
            gateway = ip_profile.props.gateway
            if not nm_routes and not gateway:
                continue
            default_table_id = ip_profile.props.route_table
            if gateway:
                routes.append(
                    _get_default_route_config(
                        gateway,
                        ip_profile.props.route_metric,
                        default_table_id,
                        iface_name,
                    )
                )
            # NM supports multiple route table in single profile:
            # https://bugzilla.redhat.com/show_bug.cgi?id=1436531 The
            # `ipv4.route-table` and `ipv6.route-table` will be the default
            # table id for static routes and auto routes. But each static route
            # can still specify route table id.
            for nm_route in nm_routes:
                table_id = _get_per_route_table_id(nm_route, default_table_id)
                route_entry = _nm_route_to_route(
                    nm_route, table_id, iface_name
                )
                if route_entry:
                    routes.append(route_entry)
    routes.sort(
        key=itemgetter(
            Route.TABLE_ID, Route.NEXT_HOP_INTERFACE, Route.DESTINATION
        )
    )
    return routes


def _get_per_route_table_id(nm_route, default_table_id):
    table = nm_route.get_attribute(NM.IP_ROUTE_ATTRIBUTE_TABLE)
    return int(table.get_uint32()) if table else default_table_id


def _nm_route_to_route(nm_route, table_id, iface_name):
    dst = "{ip}/{prefix}".format(
        ip=nm_route.get_dest(), prefix=nm_route.get_prefix()
    )
    next_hop = nm_route.get_next_hop() or ""
    metric = int(nm_route.get_metric())

    return {
        Route.TABLE_ID: table_id,
        Route.DESTINATION: dst,
        Route.NEXT_HOP_INTERFACE: iface_name,
        Route.NEXT_HOP_ADDRESS: next_hop,
        Route.METRIC: metric,
    }


def _get_default_route_config(gateway, metric, default_table_id, iface_name):
    if iplib.is_ipv6_address(gateway):
        destination = IPV6_DEFAULT_GATEWAY_DESTINATION
    else:
        destination = IPV4_DEFAULT_GATEWAY_DESTINATION
    return {
        Route.TABLE_ID: default_table_id,
        Route.DESTINATION: destination,
        Route.NEXT_HOP_INTERFACE: iface_name,
        Route.NEXT_HOP_ADDRESS: gateway,
        Route.METRIC: metric,
    }


def add_routes(setting_ip, routes):
    for route in routes:
        _add_specfic_route(setting_ip, route)


def _add_specfic_route(setting_ip, route):
    destination, prefix_len = route[Route.DESTINATION].split("/")
    prefix_len = int(prefix_len)
    if iplib.is_ipv6_address(destination):
        family = socket.AF_INET6
    else:
        family = socket.AF_INET
    metric = route.get(Route.METRIC, Route.USE_DEFAULT_METRIC)
    next_hop = route[Route.NEXT_HOP_ADDRESS]
    ip_route = NM.IPRoute.new(
        family, destination, prefix_len, next_hop, metric
    )
    table_id = route.get(Route.TABLE_ID, Route.USE_DEFAULT_ROUTE_TABLE)
    ip_route.set_attribute(
        NM.IP_ROUTE_ATTRIBUTE_TABLE, GLib.Variant.new_uint32(table_id)
    )
    # Duplicate route entry will be ignored by libnm.
    setting_ip.add_route(ip_route)


def get_static_gateway_iface(family, iface_routes):
    """
    Return one interface with gateway for given IP family.
    Return None if not found.
    """
    destination = (
        IPV6_DEFAULT_GATEWAY_DESTINATION
        if family == Interface.IPV6
        else IPV4_DEFAULT_GATEWAY_DESTINATION
    )
    for iface_name, routes in iface_routes.items():
        for route in routes:
            if route[Route.DESTINATION] == destination:
                return iface_name
    return None


def add_route_rules(setting_ip, family, rules):
    for rule in rules:
        setting_ip.add_routing_rule(_rule_info_to_nm_rule(rule, family))


def _rule_info_to_nm_rule(rule, family):
    nm_rule = NM.IPRoutingRule.new(family)
    ip_from = rule.get(RouteRule.IP_FROM)
    ip_to = rule.get(RouteRule.IP_TO)
    if not ip_from and not ip_to:
        raise NmstateValueError(
            f"Neither {RouteRule.IP_FROM} or {RouteRule.IP_TO} is defined"
        )

    if ip_from:
        nm_rule.set_from(*iplib.ip_address_full_to_tuple(ip_from))
    if ip_to:
        nm_rule.set_to(*iplib.ip_address_full_to_tuple(ip_to))

    priority = rule.get(RouteRule.PRIORITY)
    if priority and priority != RouteRule.USE_DEFAULT_PRIORITY:
        nm_rule.set_priority(priority)
    else:
        nm_rule.set_priority(ROUTE_RULE_DEFAULT_PRIORIRY)
    table = rule.get(RouteRule.ROUTE_TABLE)
    if table and table != RouteRule.USE_DEFAULT_ROUTE_TABLE:
        nm_rule.set_table(table)
    else:
        nm_rule.set_table(iplib.KERNEL_MAIN_ROUTE_TABLE_ID)
    return nm_rule


def get_route_rules_from_nm_profiles(nm_profiles):
    rules = []
    for nm_profile in nm_profiles:
        for ip_profile in (
            nm_profile.get_setting_ip6_config(),
            nm_profile.get_setting_ip4_config(),
        ):
            if ip_profile:
                for i in range(ip_profile.get_num_routing_rules()):
                    nm_rule = ip_profile.get_routing_rule(i)
                    rules.append(_nm_rule_to_info(nm_rule))
    return rules


def _nm_rule_to_info(nm_rule):
    info = {
        RouteRule.IP_FROM: _nm_rule_get_from(nm_rule),
        RouteRule.IP_TO: _nm_rule_get_to(nm_rule),
        RouteRule.PRIORITY: nm_rule.get_priority(),
        RouteRule.ROUTE_TABLE: nm_rule.get_table(),
    }
    cleanup_keys = [key for key, val in info.items() if val is None]
    for key in cleanup_keys:
        del info[key]

    return info


def _nm_rule_get_from(nm_rule):
    if nm_rule.get_from():
        return iplib.to_ip_address_full(
            nm_rule.get_from(), nm_rule.get_from_len()
        )
    return None


def _nm_rule_get_to(nm_rule):
    if nm_rule.get_to():
        return iplib.to_ip_address_full(nm_rule.get_to(), nm_rule.get_to_len())
    return None



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from abc import ABCMeta
from abc import abstractproperty
from abc import abstractmethod

from .error import NmstatePluginError


class NmstatePlugin(metaclass=ABCMeta):
    OVS_CAPABILITY = "openvswitch"
    TEAM_CAPABILITY = "team"

    PLUGIN_CAPABILITY_IFACE = "interface"
    PLUGIN_CAPABILITY_ROUTE = "route"
    PLUGIN_CAPABILITY_ROUTE_RULE = "route_rule"
    PLUGIN_CAPABILITY_DNS = "dns"

    DEFAULT_PRIORITY = 10

    def unload(self):
        pass

    @property
    def checkpoint(self):
        return None

    def refresh_content(self):
        pass

    @abstractproperty
    def name(self):
        pass

    @property
    def priority(self):
        return NmstatePlugin.DEFAULT_PRIORITY

    def get_interfaces(self):
        """
        Return a list of dict with network interface running status with
        mix of running status and running configure.
        """
        raise NmstatePluginError(
            f"Plugin {self.name} BUG: get_interfaces() not implemented"
        )

    def get_running_config_interfaces(self):
        """
        Return a list of dict with network interface running configuration.
        Notes:
            * the IP/DHCP/Route retrieved from DHCP/Autoconf are not running
              configuration.
        """
        return []

    def get_saved_config_interfaces(self):
        """
        Return a list of dict with network interface saved(persistent after
        reboot) configuration.
        Notes:
            * the IP/DHCP/Route retrieved from DHCP/Autoconf are not saved
              configuration.
            * Memory only configuration is not saved configration.
        """
        return []

    def apply_changes(self, net_state, save_to_disk):
        pass

    @property
    def capabilities(self):
        return []

    @abstractmethod
    def plugin_capabilities(self):
        pass

    def create_checkpoint(self, timeout):
        return None

    def rollback_checkpoint(self, checkpoint=None):
        pass

    def destroy_checkpoint(self, checkpoint=None):
        pass

    def get_routes(self):
        raise NmstatePluginError(
            f"Plugin {self.name} BUG: get_routes() not implemented"
        )

    def get_saved_routes(self):
        return []

    def get_route_rules(self):
        raise NmstatePluginError(
            f"Plugin {self.name} BUG: get_route_rules() not implemented"
        )

    def get_saved_route_rules(self):
        return []

    def get_dns_client_config(self):
        raise NmstatePluginError(
            f"Plugin {self.name} BUG: get_dns_client_config() not implemented"
        )

    def get_global_state(self):
        """
        Allowing plugin to append global information to content of
        `libnmstate.show()`.
        """
        return {}

    def get_saved_dns_client_config(self):
        return {}

    @property
    def is_supplemental_only(self):
        """
        Return True when plugin can only append information to
        interfaces reported by other plugin.
        Retrun False when plugin can report new interface.
        """
        return False



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import InterfaceType
from libnmstate.schema import MacVtap

from .macvlan import NisporPluginMacVlanIface


MACVTAP_MODES = {
    "unknown": MacVtap.Mode.UNKNOWN,
    "vepa": MacVtap.Mode.VEPA,
    "bridge": MacVtap.Mode.BRIDGE,
    "private": MacVtap.Mode.PRIVATE,
    "passthru": MacVtap.Mode.PASSTHRU,
    "source": MacVtap.Mode.SOURCE,
}


class NisporPluginMacVtapIface(NisporPluginMacVlanIface):
    @property
    def type(self):
        return InterfaceType.MAC_VTAP

    def to_dict(self, config_only):
        info = super().to_dict(config_only)
        info[MacVtap.CONFIG_SUBTREE] = {
            MacVtap.BASE_IFACE: self._np_iface.base_iface,
            MacVtap.MODE: MACVTAP_MODES.get(
                self._np_iface.mode, MacVtap.Mode.UNKNOWN
            ),
            MacVtap.PROMISCUOUS: self._flag_to_promiscuous(
                self._np_iface.mac_vlan_flags
            ),
        }

        return info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import InterfaceType

from libnmstate.ifaces.dummy import DummyIface

from ..testlib.ifacelib import gen_foo_iface_info


class TestDummyIface:
    def _gen_iface_info(self):
        return gen_foo_iface_info(iface_type=InterfaceType.DUMMY)

    def test_team_is_virtual(self):
        assert DummyIface(self._gen_iface_info()).is_virtual



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy
import os
import tempfile

import pytest

import libnmstate
from libnmstate.error import NmstateDependencyError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import DNS
from libnmstate.schema import Route
from libnmstate.schema import RouteRule
from libnmstate.plugin import NmstatePlugin

from .testlib import statelib
from .testlib.servicelib import disable_service


FOO_IFACE_NAME = "foo1"
FOO_IFACE_STATES = [
    {
        Interface.NAME: FOO_IFACE_NAME,
        Interface.TYPE: InterfaceType.OTHER,
        Interface.STATE: InterfaceState.UP,
        "foo": {"a": 1, "b": 2},
    }
]

BAR_IFACE_NAME = "bar1"
BAR_IFACE_STATES = [
    {
        Interface.NAME: BAR_IFACE_NAME,
        Interface.TYPE: InterfaceType.OTHER,
        Interface.STATE: InterfaceState.UP,
        "foo": {"a": 2},
    },
    {
        Interface.NAME: FOO_IFACE_NAME,
        Interface.TYPE: InterfaceType.OTHER,
        Interface.STATE: InterfaceState.UP,
        "foo": {"a": 3},
    },
]

TEST_ROUTE_ENTRY = {
    Route.DESTINATION: "198.51.100.0/24",
    Route.METRIC: 103,
    Route.NEXT_HOP_ADDRESS: "192.0.2.1",
    Route.NEXT_HOP_INTERFACE: FOO_IFACE_NAME,
    Route.TABLE_ID: 100,
}

TEST_ROUTE_STATE = {
    Route.RUNNING: [TEST_ROUTE_ENTRY],
    Route.CONFIG: [],
}

TEST_DNS_STATE = {
    DNS.RUNNING: {
        DNS.SERVER: ["2001:4860:4860::8888", "1.1.1.1"],
        DNS.SEARCH: ["example.org", "example.com"],
    },
    DNS.CONFIG: [],
}

TEST_ROUTE_RULE_STATE = {
    RouteRule.CONFIG: [
        {
            RouteRule.IP_FROM: "2001:db8:a::/64",
            RouteRule.IP_TO: "2001:db8:f::/64",
            RouteRule.PRIORITY: 1000,
            RouteRule.ROUTE_TABLE: 100,
        },
        {
            RouteRule.IP_FROM: "203.0.113.0/24",
            RouteRule.IP_TO: "192.0.2.0/24",
            RouteRule.PRIORITY: 1001,
            RouteRule.ROUTE_TABLE: 101,
        },
    ]
}

GET_IFACES_FORMAT = """
    def get_interfaces(self):
        return {ifaces}
"""

GET_ROUTES_FORMAT = """
    def get_routes(self):
        return {routes}
"""

GET_ROUTE_RULES_FORMAT = """
    def get_route_rules(self):
        return {route_rules}
"""

GET_DNS_FORMAT = """
    def get_dns_client_config(self):
        return {dns_config}
"""

LO_IFACE_INFO = {
    Interface.NAME: "lo",
    Interface.TYPE: InterfaceType.UNKNOWN,
    Interface.STATE: InterfaceState.UP,
    Interface.IPV4: {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.ADDRESS: [
            {
                InterfaceIPv4.ADDRESS_IP: "127.0.0.1",
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 8,
            }
        ],
    },
    Interface.IPV6: {
        InterfaceIPv6.ENABLED: True,
        InterfaceIPv6.ADDRESS: [
            {
                InterfaceIPv6.ADDRESS_IP: "::1",
                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 128,
            }
        ],
    },
    Interface.MAC: "00:00:00:00:00:00",
    Interface.MTU: 65536,
}


@pytest.fixture
def tmp_plugin_dir():
    with tempfile.TemporaryDirectory() as plugin_dir:
        os.environ["NMSTATE_PLUGIN_DIR"] = plugin_dir
        yield plugin_dir
        os.environ.pop("NMSTATE_PLUGIN_DIR")


@pytest.fixture
def with_foo_plugin(tmp_plugin_dir):
    _gen_plugin_foo(tmp_plugin_dir)


@pytest.fixture
def with_multiple_plugins(tmp_plugin_dir):
    _gen_plugin_foo(tmp_plugin_dir)
    _gen_plugin_bar(tmp_plugin_dir)


@pytest.fixture
def with_route_plugin(tmp_plugin_dir):
    _gen_plugin_route_foo(tmp_plugin_dir)


@pytest.fixture
def with_route_rule_plugin(tmp_plugin_dir):
    _gen_plugin_route_rule_foo(tmp_plugin_dir)


@pytest.fixture
def with_dns_plugin(tmp_plugin_dir):
    _gen_plugin_dns_foo(tmp_plugin_dir)


def _gen_plugin(
    plugin_dir,
    plugin_name,
    plugin_class,
    priority,
    ifaces=None,
    routes=None,
    dns_config=None,
    route_rules=None,
):
    plugin_capabilities = []
    get_funs_txt = ""
    if ifaces:
        plugin_capabilities.append(NmstatePlugin.PLUGIN_CAPABILITY_IFACE)
        get_funs_txt += GET_IFACES_FORMAT.format(ifaces=f"{ifaces}")
    if routes:
        plugin_capabilities.append(NmstatePlugin.PLUGIN_CAPABILITY_ROUTE)
        get_funs_txt += GET_ROUTES_FORMAT.format(routes=f"{routes}")
    if route_rules:
        plugin_capabilities.append(NmstatePlugin.PLUGIN_CAPABILITY_ROUTE_RULE)
        get_funs_txt += GET_ROUTE_RULES_FORMAT.format(
            route_rules=f"{route_rules}"
        )
    if dns_config:
        plugin_capabilities.append(NmstatePlugin.PLUGIN_CAPABILITY_DNS)
        get_funs_txt += GET_DNS_FORMAT.format(dns_config=f"{dns_config}")

    plugin_txt = f"""
from libnmstate.plugin import NmstatePlugin
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState

class {plugin_class}(NmstatePlugin):
    @property
    def name(self):
        return "{plugin_name}"

    @property
    def priority(self):
        return {priority}

    @property
    def plugin_capabilities(self):
        return {plugin_capabilities}

{get_funs_txt}

NMSTATE_PLUGIN = {plugin_class}
"""
    with open(f"{plugin_dir}/nmstate_plugin_{plugin_name}.py", "w") as fd:
        fd.write(plugin_txt)


def _gen_plugin_foo(plugin_dir):
    _gen_plugin(
        plugin_dir,
        "foo",
        "NmstateFooPlugin",
        NmstatePlugin.DEFAULT_PRIORITY + 1,
        ifaces=FOO_IFACE_STATES,
    )


def _gen_plugin_bar(plugin_dir):
    _gen_plugin(
        plugin_dir,
        "bar",
        "NmstateBarPlugin",
        NmstatePlugin.DEFAULT_PRIORITY + 2,
        ifaces=BAR_IFACE_STATES,
    )


def _gen_plugin_route_foo(plugin_dir):
    _gen_plugin(
        plugin_dir,
        "route_foo",
        "NmstateRouteFooPlugin",
        NmstatePlugin.DEFAULT_PRIORITY + 1,
        routes=TEST_ROUTE_STATE,
    )


def _gen_plugin_dns_foo(plugin_dir):
    _gen_plugin(
        plugin_dir,
        "dns_foo",
        "NmstateDnsFooPlugin",
        NmstatePlugin.DEFAULT_PRIORITY + 1,
        dns_config=TEST_DNS_STATE,
    )


def _gen_plugin_route_rule_foo(plugin_dir):
    _gen_plugin(
        plugin_dir,
        "route_rule_foo",
        "NmstateRouteRuleFooPlugin",
        NmstatePlugin.DEFAULT_PRIORITY + 1,
        route_rules=TEST_ROUTE_RULE_STATE,
    )


def test_load_foo_plugin(with_foo_plugin):
    current_state = statelib.show_only((FOO_IFACE_NAME,))
    assert current_state[Interface.KEY] == FOO_IFACE_STATES


def test_two_plugins_with_merged_iface_by_priority(with_multiple_plugins):
    current_state = statelib.show_only((BAR_IFACE_NAME, FOO_IFACE_NAME))
    expected_ifaces = copy.deepcopy(BAR_IFACE_STATES)
    expected_ifaces[1]["foo"]["b"] = 2
    assert current_state[Interface.KEY] == expected_ifaces


def test_load_external_route_plugin(with_route_plugin):
    state = libnmstate.show()
    assert TEST_ROUTE_ENTRY in state[Route.KEY][Route.RUNNING]


def test_load_external_route_rule_plugin(with_route_rule_plugin):
    state = libnmstate.show()
    assert state[RouteRule.KEY] == TEST_ROUTE_RULE_STATE


def test_load_external_dns_plugin(with_dns_plugin):
    state = libnmstate.show()
    assert state[DNS.KEY] == TEST_DNS_STATE


@pytest.fixture
def stop_nm_service():
    with disable_service("NetworkManager"):
        yield


def test_network_manager_plugin_with_daemon_stopped(stop_nm_service):
    with pytest.raises(NmstateDependencyError):
        from libnmstate.nm import NetworkManagerPlugin

        NetworkManagerPlugin()

    state = statelib.show_only(("lo",))
    assert state[Interface.KEY][0] == LO_IFACE_INFO



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import re
import subprocess

RC_SUCCESS = 0
RC_FAIL2 = 2


def exec_cmd(cmd, env=None, stdin=None, check=False):
    """
    Execute cmd in an external process, collect its output and returncode

    :param cmd: an iterator of strings to be passed as exec(2)'s argv
    :param env: an optional dictionary to be placed as environment variables
                of the external process. If None, the environment of the
                calling process is used.
    :returns: a 3-tuple of the process's
              (returncode, stdout content as unicode,
               stderr content as unicode.)
    """
    logging.debug(command_log_line(cmd))

    p = subprocess.run(
        cmd,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
        check=check,
        input=stdin,
    )

    out = p.stdout
    err = p.stderr

    logging.debug(_retcode_log_line(p.returncode, err=err))

    return (p.returncode, out.decode("utf-8"), err.decode("utf-8"))


def command_log_line(args, cwd=None):
    return "{0} (cwd {1})".format(_list2cmdline(args), cwd)


def format_exec_cmd_result(result):
    return "rc={}, out={}, err={}".format(*result)


def _retcode_log_line(code, err=None):
    result = "SUCCESS" if code == 0 else "FAILED"
    return "{0}: <err> = {1!r}; <rc> = {2!r}".format(result, err, code)


def _list2cmdline(args):
    """
    Convert argument list for exec_cmd to string for logging.
    The purpose of this log is to make it easy to run commands in the shell for
    debugging.
    """
    parts = []
    for arg in args:
        if _needs_quoting(arg) or arg == "":
            arg = "'" + arg.replace("'", r"'\''") + "'"
        parts.append(arg)
    return " ".join(parts)


# This function returns truthy value if its argument contains unsafe characters
# for including in a command passed to the shell. The safe characters were
# stolen from pipes._safechars.
_needs_quoting = re.compile(r"[^A-Za-z0-9_%+,\-./:=@]").search



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os

import pytest

import libnmstate
from libnmstate.error import NmstateValueError
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import InfiniBand
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge

from .testlib import assertlib
from .testlib import statelib
from .testlib.bondlib import bond_interface
from .testlib.bridgelib import linux_bridge
from .testlib.bridgelib import add_port_to_bridge


TEST_BRIDGE0 = "linux-br0"
TEST_BOND99 = "bond99"
TEST_PKEY1 = "0x80fe"
TEST_PKEY2 = "0x80ff"

IPV4_ADDRESS1 = "192.0.2.251"
IPV6_ADDRESS1 = "2001:db8:1::1"

TEST_STATIC_IPV4_CONFIG = {
    InterfaceIPv4.ENABLED: True,
    InterfaceIPv4.DHCP: False,
    InterfaceIPv4.ADDRESS: [
        {
            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
        }
    ],
}

TEST_STATIC_IPV6_CONFIG = {
    InterfaceIPv6.ENABLED: True,
    InterfaceIPv6.DHCP: False,
    InterfaceIPv6.AUTOCONF: False,
    InterfaceIPv6.ADDRESS: [
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        }
    ],
}


def _test_nic_name():
    return os.environ.get("TEST_REAL_NIC")


def _test_ib_mode():
    return (
        InfiniBand.Mode.CONNECTED
        if os.environ.get("TEST_IB_CONNECTED_MODE")
        else InfiniBand.Mode.DATAGRAM
    )


def _gen_ib_iface_info(
    base_iface,
    pkey=InfiniBand.DEFAULT_PKEY,
    mode=InfiniBand.Mode.DATAGRAM,
):
    if pkey == InfiniBand.DEFAULT_PKEY:
        iface_name = base_iface
        base_iface = ""
    else:
        iface_name = f"{base_iface}.{pkey[2:]}"

    return {
        Interface.NAME: iface_name,
        Interface.TYPE: InterfaceType.INFINIBAND,
        InfiniBand.CONFIG_SUBTREE: {
            InfiniBand.PKEY: pkey,
            InfiniBand.MODE: mode,
            InfiniBand.BASE_IFACE: base_iface,
        },
    }


@pytest.fixture
def ib_base_nic():
    iface_info = _gen_ib_iface_info(_test_nic_name(), mode=_test_ib_mode())
    libnmstate.apply({Interface.KEY: [iface_info]})
    yield iface_info
    iface_info[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply(
        {Interface.KEY: [iface_info]},
        verify_change=False,
    )


@pytest.fixture
def ib_pkey_nic1(ib_base_nic):
    iface_info = _gen_ib_iface_info(
        _test_nic_name(), TEST_PKEY1, _test_ib_mode()
    )
    libnmstate.apply({Interface.KEY: [iface_info]})
    yield iface_info
    iface_info[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply(
        {Interface.KEY: [iface_info]},
        verify_change=False,
    )


@pytest.fixture
def ib_pkey_nic2(ib_base_nic):
    iface_info = _gen_ib_iface_info(
        _test_nic_name(), TEST_PKEY2, _test_ib_mode()
    )
    libnmstate.apply({Interface.KEY: [iface_info]})
    yield iface_info
    iface_info[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply(
        {Interface.KEY: [iface_info]},
        verify_change=False,
    )


@pytest.fixture
def empty_test_bridge():
    bridge_config = {}
    with linux_bridge(TEST_BRIDGE0, bridge_config) as state:
        yield state


@pytest.fixture
def empty_test_bond():
    with bond_interface(TEST_BOND99, []) as state:
        yield state


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC(TEST_IB_CONNECTED_MODE optionally) "
    "for infiniband test",
)
class TestInfiniBand:
    def test_create_and_remove_pkey_nic(self, ib_base_nic):
        desired_state = {
            Interface.KEY: [
                _gen_ib_iface_info(
                    _test_nic_name(), TEST_PKEY1, _test_ib_mode()
                )
            ]
        }
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            desired_state[Interface.KEY][0][
                Interface.STATE
            ] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

    def test_change_mtu_of_base_nic(self, ib_base_nic):
        desired_state = {
            Interface.KEY: [
                {Interface.NAME: _test_nic_name(), Interface.MTU: 1280}
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_change_mtu_of_pkey_nic(self, ib_pkey_nic1):
        iface_info = ib_pkey_nic1
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: iface_info[Interface.NAME],
                        Interface.MTU: 1280,
                    }
                ]
            }
        )

        iface_info[Interface.MTU] = 1280
        desired_state = {Interface.KEY: [iface_info]}
        assertlib.assert_state_match(desired_state)

    def test_remove_base_nic_got_pkey_nics_removed_automatically(
        self, ib_pkey_nic1
    ):
        pkey_iface_name = ib_pkey_nic1[Interface.NAME]
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: _test_nic_name(),
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )
        current_state = statelib.show_only((pkey_iface_name,))
        if len(current_state[Interface.KEY]):
            # The interface might be deleted after base interface
            cur_pkey_iface_info = current_state[Interface.KEY][0]
            assert cur_pkey_iface_info[Interface.STATE] == InterfaceState.DOWN

    def test_change_ip_of_base_nic(self, ib_base_nic):
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: _test_nic_name(),
                    Interface.IPV4: TEST_STATIC_IPV4_CONFIG,
                    Interface.IPV6: TEST_STATIC_IPV6_CONFIG,
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_change_ip_of_pkey_nic(self, ib_pkey_nic1):
        iface_name = ib_pkey_nic1[Interface.NAME]
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: iface_name,
                    Interface.IPV4: TEST_STATIC_IPV4_CONFIG,
                    Interface.IPV6: TEST_STATIC_IPV6_CONFIG,
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_add_base_nic_to_linux_bridge(
        self, empty_test_bridge, ib_base_nic
    ):
        """
        The InfiniBand is IP over IB, hence there is ethernet layer for
        IPoIB NIC, so there is no way for adding a IPoIB NIC to bridge
        """
        bridge_iface_info = empty_test_bridge[Interface.KEY][0]
        bridge_iface_info[LinuxBridge.CONFIG_SUBTREE] = {}
        add_port_to_bridge(
            bridge_iface_info[LinuxBridge.CONFIG_SUBTREE], _test_nic_name()
        )
        desired_state = {Interface.KEY: [bridge_iface_info]}

        with pytest.raises(NmstateValueError):
            libnmstate.apply(desired_state)

    def test_add_pkey_nic_to_linux_bridge(
        self, empty_test_bridge, ib_pkey_nic1
    ):
        pkey_iface_name = ib_pkey_nic1[Interface.NAME]
        bridge_iface_info = empty_test_bridge[Interface.KEY][0]
        bridge_iface_info[LinuxBridge.CONFIG_SUBTREE] = {}
        add_port_to_bridge(
            bridge_iface_info[LinuxBridge.CONFIG_SUBTREE], pkey_iface_name
        )
        desired_state = {Interface.KEY: [bridge_iface_info]}

        with pytest.raises(NmstateValueError):
            libnmstate.apply(desired_state)

    def test_add_base_nic_to_active_backup_bond(
        self, empty_test_bond, ib_base_nic
    ):
        with bond_interface(
            TEST_BOND99, [_test_nic_name()], create=False
        ) as desired_state:
            desired_state[Interface.KEY][0][Bond.CONFIG_SUBTREE][
                Bond.MODE
            ] = BondMode.ACTIVE_BACKUP
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)

    def test_add_pkey_nic_to_active_backup_bond(
        self,
        empty_test_bond,
        ib_pkey_nic1,
        ib_pkey_nic2,
    ):
        port1_name = ib_pkey_nic1[Interface.NAME]
        port2_name = ib_pkey_nic2[Interface.NAME]
        with bond_interface(
            TEST_BOND99, [port1_name, port2_name], create=False
        ) as desired_state:
            desired_state[Interface.KEY][0][Bond.CONFIG_SUBTREE][
                Bond.MODE
            ] = BondMode.ACTIVE_BACKUP
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)

    def test_expect_exception_when_adding_base_nic_to_round_robin_bond(
        self, empty_test_bond, ib_base_nic
    ):
        """
        IP over InfiniBand interface is only allowd to be port of active-backup
        bond.
        """
        with bond_interface(
            TEST_BOND99, [_test_nic_name()], create=False
        ) as desired_state:
            desired_state[Interface.KEY][0][Bond.CONFIG_SUBTREE][
                Bond.MODE
            ] = BondMode.ROUND_ROBIN
            with pytest.raises(NmstateValueError):
                libnmstate.apply(desired_state)

    def test_expect_exception_when_adding_pkey_nic_to_round_robin_bond(
        self,
        empty_test_bond,
        ib_pkey_nic1,
        ib_pkey_nic2,
    ):
        """
        IP over InfiniBand interface is only allowd to be port of active-backup
        bond.
        """
        port1_name = ib_pkey_nic1[Interface.NAME]
        port2_name = ib_pkey_nic2[Interface.NAME]
        with bond_interface(
            TEST_BOND99, [port1_name, port2_name], create=False
        ) as desired_state:
            desired_state[Interface.KEY][0][Bond.CONFIG_SUBTREE][
                Bond.MODE
            ] = BondMode.ROUND_ROBIN
            with pytest.raises(NmstateValueError):
                libnmstate.apply(desired_state)

    def test_show_saved_config_with_pkey_nic_down(self, ib_pkey_nic1):
        iface_name = ib_pkey_nic1[Interface.NAME]
        running_state = statelib.show_only((iface_name,))
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: iface_name,
                        Interface.STATE: InterfaceState.DOWN,
                    }
                ]
            }
        )
        saved_state = statelib.show_saved_config_only((iface_name,))

        assert (
            saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
        )
        assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.ifaces import Ifaces
from libnmstate.schema import VLAN
from libnmstate.schema import InfiniBand
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.vlan import VlanIface

from ..testlib.ifacelib import gen_foo_iface_info

BASE_IFACE_NAME = "base1"


class TestVlanIface:
    def _gen_base_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.ETHERNET)
        iface_info[Interface.NAME] = BASE_IFACE_NAME
        iface_info[Interface.MTU] = 1500

        return iface_info

    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.VLAN)
        iface_info[VLAN.CONFIG_SUBTREE] = {
            VLAN.ID: 101,
            VLAN.BASE_IFACE: BASE_IFACE_NAME,
        }
        return iface_info

    def test_get_parent(self):
        assert VlanIface(self._gen_iface_info()).parent == BASE_IFACE_NAME

    def test_need_parent(self):
        assert VlanIface(self._gen_iface_info()).need_parent

    def test_is_virtual(self):
        assert VlanIface(self._gen_iface_info()).is_virtual

    def test_can_have_ip_as_port(self):
        assert not VlanIface(self._gen_iface_info()).can_have_ip_as_port

    def test_validate_base_iface_missing(self):
        iface_info = self._gen_iface_info()
        iface_info[VLAN.CONFIG_SUBTREE].pop(VLAN.BASE_IFACE)

        iface = VlanIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_add_vlan_with_mtu_greater_than_base(self):
        base_iface_info = self._gen_base_iface_info()
        vlan_iface_info = self._gen_iface_info()
        vlan_iface_info[Interface.MTU] = base_iface_info[Interface.MTU] + 1

        with pytest.raises(NmstateValueError):
            Ifaces(
                des_iface_infos=[base_iface_info, vlan_iface_info],
                cur_iface_infos=[],
            )

    def test_add_vlan_with_base_mtu_undefined(self):
        base_iface_info = self._gen_base_iface_info()
        base_iface_info[Interface.MTU] = None
        vlan_iface_info = self._gen_iface_info()
        vlan_iface_info[Interface.MTU] = 1501

        ifaces = Ifaces(
            des_iface_infos=[base_iface_info, vlan_iface_info],
            cur_iface_infos=[],
        )

        base_iface = ifaces.all_kernel_ifaces.get(
            base_iface_info.get(Interface.NAME)
        )

        assert base_iface.mtu == vlan_iface_info[Interface.MTU]

    def test_add_vlan_with_base_iface_infiniband(self):
        base_iface_info = gen_foo_iface_info(
            iface_type=InterfaceType.INFINIBAND
        )
        base_iface_info[Interface.NAME] = "ib0"
        base_iface_info[Interface.MTU] = 1500
        base_iface_info[InfiniBand.CONFIG_SUBTREE] = {
            InfiniBand.PKEY: InfiniBand.DEFAULT_PKEY,
            InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
        }

        vlan_iface_info = self._gen_iface_info()
        vlan_iface_info[VLAN.CONFIG_SUBTREE][VLAN.BASE_IFACE] = "ib0"

        with pytest.raises(NmstateValueError):
            Ifaces(
                des_iface_infos=[base_iface_info, vlan_iface_info],
                cur_iface_infos=[],
            )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import LinuxBridge
from libnmstate.schema import VLAN


@pytest.mark.tier1
@pytest.mark.slow
def test_lot_of_vlans_with_bridges(eth1_up):
    interfaces = []
    for i in range(100, 400):
        interfaces.extend(
            [
                {
                    Interface.NAME: "vlan." + str(i),
                    Interface.TYPE: InterfaceType.VLAN,
                    Interface.STATE: InterfaceState.UP,
                    VLAN.CONFIG_SUBTREE: {VLAN.BASE_IFACE: "eth1", VLAN.ID: i},
                },
                {
                    Interface.NAME: "linux-br" + str(i),
                    Interface.TYPE: InterfaceType.LINUX_BRIDGE,
                    Interface.STATE: InterfaceState.UP,
                    LinuxBridge.CONFIG_SUBTREE: {
                        LinuxBridge.PORT_SUBTREE: [
                            {LinuxBridge.Port.NAME: "vlan." + str(i)}
                        ]
                    },
                },
            ]
        )
    try:
        libnmstate.apply({Interface.KEY: interfaces})
    finally:
        for iface in interfaces:
            iface[Interface.STATE] = InterfaceState.ABSENT
        libnmstate.apply({Interface.KEY: interfaces})



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy

from libnmstate.schema import Bond
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType

from .common import NM


class Api2Nm:
    _iface_types_map = None

    @staticmethod
    def get_iface_type(name):
        return Api2Nm.get_iface_type_map().get(name, InterfaceType.UNKNOWN)

    @staticmethod
    def get_iface_type_map():
        if Api2Nm._iface_types_map is None:
            ovs_bridge = InterfaceType.OVS_BRIDGE
            ovs_interface = InterfaceType.OVS_INTERFACE
            Api2Nm._iface_types_map = {
                InterfaceType.ETHERNET: NM.SETTING_WIRED_SETTING_NAME,
                InterfaceType.BOND: NM.SETTING_BOND_SETTING_NAME,
                InterfaceType.DUMMY: NM.SETTING_DUMMY_SETTING_NAME,
                InterfaceType.TEAM: NM.SETTING_TEAM_SETTING_NAME,
                InterfaceType.VLAN: NM.SETTING_VLAN_SETTING_NAME,
                InterfaceType.VXLAN: NM.SETTING_VXLAN_SETTING_NAME,
                InterfaceType.LINUX_BRIDGE: NM.SETTING_BRIDGE_SETTING_NAME,
                InterfaceType.VRF: NM.SETTING_VRF_SETTING_NAME,
                InterfaceType.INFINIBAND: NM.SETTING_INFINIBAND_SETTING_NAME,
                InterfaceType.MAC_VTAP: NM.SETTING_MACVLAN_SETTING_NAME,
                InterfaceType.MAC_VLAN: NM.SETTING_MACVLAN_SETTING_NAME,
            }
            try:
                ovs_types = {
                    ovs_bridge: NM.SETTING_OVS_BRIDGE_SETTING_NAME,
                    InterfaceType.OVS_PORT: NM.SETTING_OVS_PORT_SETTING_NAME,
                    ovs_interface: NM.SETTING_OVS_INTERFACE_SETTING_NAME,
                }
                Api2Nm._iface_types_map.update(ovs_types)
            except AttributeError:
                pass

        return Api2Nm._iface_types_map

    @staticmethod
    def get_bond_options(iface_desired_state):
        iface_type = Api2Nm.get_iface_type(iface_desired_state[Interface.TYPE])
        if iface_type == InterfaceType.BOND:
            # Is the mode a must config parameter?
            bond_conf = iface_desired_state[Bond.CONFIG_SUBTREE]
            bond_opts = {Bond.MODE: bond_conf[Bond.MODE]}
            bond_opts.update(bond_conf.get(Bond.OPTIONS_SUBTREE, {}))
        else:
            bond_opts = {}

        return bond_opts


class Nm2Api:
    _iface_types_map = None

    @staticmethod
    def get_common_device_info(devinfo):
        type_name = devinfo["type_name"]
        if type_name != InterfaceType.ETHERNET:
            type_name = Nm2Api.get_iface_type(type_name)
        return {
            Interface.NAME: devinfo[Interface.NAME],
            Interface.TYPE: type_name,
            Interface.STATE: Nm2Api.get_iface_admin_state(
                devinfo[Interface.STATE]
            ),
        }

    @staticmethod
    def get_bond_info(bondinfo):
        bond_options = copy.deepcopy(bondinfo.get(Bond.OPTIONS_SUBTREE))
        if not bond_options:
            return {}
        bond_port = bondinfo[Bond.PORT]

        bond_mode = bond_options[Bond.MODE]
        del bond_options[Bond.MODE]
        return {
            Bond.CONFIG_SUBTREE: {
                Bond.MODE: bond_mode,
                Bond.PORT: [port.props.interface for port in bond_port],
                Bond.OPTIONS_SUBTREE: bond_options,
            }
        }

    @staticmethod
    def get_iface_type(name):
        if Nm2Api._iface_types_map is None:
            Nm2Api._iface_types_map = Nm2Api._swap_dict_keyval(
                Api2Nm.get_iface_type_map()
            )
        return Nm2Api._iface_types_map.get(name, InterfaceType.UNKNOWN)

    @staticmethod
    def get_iface_admin_state(dev_state):
        if NM.DeviceState.IP_CONFIG <= dev_state <= NM.DeviceState.ACTIVATED:
            return InterfaceState.UP
        return InterfaceState.DOWN

    @staticmethod
    def _swap_dict_keyval(dictionary):
        return {val: key for key, val in dictionary.items()}
